# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Set up/configure logging to STDERR, files and/or syslog
"""

from __future__ import annotations

import logging
import logging.handlers
from pathlib import PosixPath
from sys import stderr
from typing import Callable
from socket import SOCK_STREAM, SOCK_DGRAM, gethostname

import loguru
from loguru._defaults import (
    LOGURU_DEBUG_NO,
    LOGURU_ERROR_NO,
    LOGURU_INFO_NO,
    LOGURU_SUCCESS_NO,
    LOGURU_TRACE_NO,
    LOGURU_WARNING_NO,
)

from .settings.loggers import LogFile, Syslog


class LogManager:
    """
    Set up/configure logging to STDERR, a file and/or syslog
    """

    def __init__(self, verbosity: int):
        """
        Create the new log object
        """
        # Set up the default STDERR log
        self.setup_stderr(verbosity=verbosity)

        # Create a logging context for self
        self.log = loguru.logger.bind(
            check_name="MASTER",
            subsystem="logging",
        )

        # Log the creation of the log manager
        self.log.bind(event="debug").debug("New LogManager object created")

        # Log creation of STDERR logger
        self.log.bind(event="debug").debug(
            "STDERR logger created at level {level}",
            level=self._stderr_level,
        )
        self.log.bind(event="datadump").trace(
            "STDERR log format string:\n{log_format}",
            log_format=self._stderr_log_format,
        )

    def setup_stderr(self, verbosity: int) -> None:
        """
        Setup logging to STDERR
        """

        # Define the available log levels for verbosity flags
        levels = {
            0: LOGURU_ERROR_NO,
            1: LOGURU_WARNING_NO,
            2: LOGURU_SUCCESS_NO,
            3: LOGURU_INFO_NO,
            4: LOGURU_DEBUG_NO,
            5: LOGURU_TRACE_NO,
        }

        # Fix typing for loguru levels
        assert isinstance(LOGURU_ERROR_NO, int)
        assert isinstance(LOGURU_WARNING_NO, int)
        assert isinstance(LOGURU_SUCCESS_NO, int)
        assert isinstance(LOGURU_INFO_NO, int)
        assert isinstance(LOGURU_DEBUG_NO, int)
        assert isinstance(LOGURU_TRACE_NO, int)

        # Get log level to set; if debug is disabled, limit to info level to prevent log spam
        self._stderr_level: int = (  # type: ignore
            levels.get(verbosity, LOGURU_TRACE_NO) if __debug__ else LOGURU_INFO_NO
        )
        # Set type for MyPy
        assert isinstance(self._stderr_level, int)

        # Create the common log format for STDERR messages
        log_format: list[str] = [
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>",
            "<level>{level: <8}</level>",
            "<light-blue>PID {process: <8}</light-blue>",
            "<cyan>{extra[check_name]}</cyan>",
            "<green>{extra[subsystem]}/{extra[event]}</green>",
            "<level>{message}</level>",
        ]

        # If the log level is less than LOGURU_INFO_NO, add the function name, line number and file name to the log
        # format. If LOGURU_DEBUG_NO or less the full file path will be used
        if self._stderr_level < LOGURU_INFO_NO:
            if self._stderr_level < LOGURU_DEBUG_NO:
                log_format.insert(
                    3, "<magenta>{file.path}:{line} {function}()</magenta>"
                )
            else:
                log_format.insert(3, "<magenta>{file}:{line} {function}()</magenta>")

        # Combine the log format strings
        self._stderr_log_format = " | ".join(log_format)

        # Remove the default logger
        loguru.logger.remove()

        # Add the new logger set at the correct log level and with correct format
        loguru.logger.add(
            stderr, format=self._stderr_log_format, level=self._stderr_level
        )

    def setup(self, loggers: list[LogFile | Syslog]) -> None:
        """
        Configure the list of log files/syslog destinations
        """
        # Log start of setup
        self.log.bind(event="debug").debug("Setting up file/syslog loggers")

        # Loop through the list of loggers
        for config in loggers:
            # Set up the required logger
            self._setup_logger(config=config)

    def _setup_logger(self, config: LogFile | Syslog) -> None:
        """
        Create a single logger for a log file or syslog destination
        """
        # Log setup of the individual logger
        self.log.bind(event="debug").debug(
            "Creating {log_type} logger", log_type=config.method
        )
        self.log.bind(event="datadump").trace(
            "Logger config:\n{config}",
            config=config.pretty,
        )

        # Generate the log format
        log_format = self.log_format(config=config)
        self.log.bind(event="datadump").trace(
            "Generated log format:\n{log_format}", log_format=log_format
        )

        # Create a filter for the log
        log_filter = self.log_filter(config=config)

        # Uppercase the log level
        level = config.level.upper()

        # If the logger is a syslog destination, set the destination and create the log handler first
        if isinstance(config, Syslog):
            # First check if logging to the local syslog socket
            if config.destination == "/dev/log" or isinstance(
                config.destination, PosixPath
            ):
                # Create the log handler
                handler = logging.handlers.SysLogHandler(
                    address=f"{config.destination}"
                )
            else:
                # Set the protocol to use
                if config.protocol == "tcp":
                    protocol = SOCK_STREAM
                else:
                    protocol = SOCK_DGRAM

                # Create the log handler
                handler = logging.handlers.SysLogHandler(
                    address=(config.destination, config.port),
                    socktype=protocol,
                )

            # Add the new logger
            logger_id = loguru.logger.add(
                handler,
                format=log_format,
                level=level,
                filter=log_filter,
            )

        # Logging is to a file, set compression type and create logger
        else:
            # Set compression if enabled
            compression = config.compression_format if config.compress else None

            # Add the new logger
            logger_id = loguru.logger.add(
                config.destination,
                format=log_format,
                level=level,
                rotation=config.size,
                compression=compression,
                enqueue=True,
                filter=log_filter,
                retention=config.count,
            )

        # Set the logger ID for the config
        config.set_logger_id(logger_id)

    @staticmethod
    def log_format(config: LogFile | Syslog) -> str:
        """
        Generate the log format string for the supplied log configuration
        """
        # Return the user defined format if set
        if config.formatter:
            return config.formatter

        # Create a list of fields to include in the log format
        fields: list[tuple[str, str]] = []

        # When using file logging, the date/time is formatted uniquely depending on structured logging
        if isinstance(config, LogFile):
            if config.structured:
                fields.append(("date", "{time:YYYY-MM-DD}"))
                fields.append(("time", "{time:HH:mm:ss.SSS}"))
            else:
                fields.append(("time", "{time:YYYY-MM-DD HH:mm:ss.SSS}"))

        # Check if logging with syslog
        if isinstance(config, Syslog):

            # If using structured logging or logging to a remote server, include a time field
            if config.structured or isinstance(config.destination, str):
                fields.append(("time", "{time:YYYY-MM-DD HH:mm:ss.SSS}"))

            # If logging to a remote server, include the hostname
            if isinstance(config.destination, str):
                fields.append(("hostname", f"{gethostname()}"))

        # Define the log level field
        if config.structured:
            fields.append(("level", "{level}"))
        else:
            fields.append(("level", "{level: <8}"))

        # Add the process ID and check name
        fields.append(("pid", "{process}"))
        fields.append(("check_name", "{extra[check_name]}"))

        # If using structured logging the file path, line number and function is split
        if config.structured:
            fields.append(("file", "{file.path}"))
            fields.append(("line", "{line}"))
            fields.append(("function", "{function}"))
        else:
            # If not using structured logging and the log level is debugging/trace, log the file and line number
            if config.level in ["debug", "trace"]:
                fields.append(("file", "{file.path}:{line}"))
                fields.append(("function", "{function}"))

        # Append the event
        fields.append(("event", "{extra[event]}"))

        # Append the subsystem
        fields.append(("subsystem", "{extra[subsystem]}"))

        # Append the log message
        fields.append(("message", "{message}"))

        # Create the log format string from the fields provided; depending on if structured formatting is used or not
        if config.structured:
            log_format = " ".join([f'"{field[0]}"="{field[1]}"' for field in fields])
        else:
            log_format = " | ".join([field[1] for field in fields])

        # Return the log format
        return log_format

    @staticmethod
    def log_filter(config: LogFile | Syslog) -> Callable:
        """
        Generate a filter for the supplied log configuration
        """

        # Create set of events, subsystems and check names to log
        events = set(config.events)
        subsystems = set(config.subsystems)
        if config.checks:
            checks = set(config.checks)
        else:
            checks = set()

        # If the log is structured or to syslog, data dumping is not permitted
        if config.structured or isinstance(config, Syslog):
            if "datadump" in events:
                events.remove("datadump")

        # Define the filter function
        def _filter(record: loguru.Record) -> bool:
            """
            Create a filter for log records based on the supplied configuration
            """
            # If filter by check name is set, ensure the record check name is in the allow list
            if checks and record["extra"]["check_name"] not in checks:
                return False

            # Ensure the record event is in the allow list
            if (
                record["extra"]["event"] not in events
                or record["extra"]["subsystem"] not in subsystems
            ):
                return False

            # If not matched by previous checks, log the message
            return True

        # Return the filter function
        return _filter



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Store configuration for an individual health check
"""

import re

from ipaddress import (
    IPv4Address,
    IPv4Network,
    IPv6Address,
    IPv6Network,
)
from pathlib import Path
from typing import Literal, Optional

from pydantic import (
    Field,
    model_validator,
    field_validator,
    IPvAnyNetwork,
    IPvAnyAddress,
    PositiveInt,
    PositiveFloat,
    NonNegativeInt,
)
from pydantic_core.core_schema import ValidationInfo
from typing_extensions import Annotated

from ._base import Base
from .checkargs.dnsargs import DNSArgs
from .checkargs.fileargs import FileArgs
from .checkargs.httpargs import HTTPArgs
from .checkargs.icmpargs import ICMPArgs
from .checkargs.ntpargs import NTPArgs
from .checkargs.shellargs import ShellArgs
from .checkargs.tcpargs import TCPArgs


class Check(Base):
    """
    The common base check method
    """

    name: str = Field(
        title="Check Name",
        description="The name of the check",
        pattern=r"^[^\"']+$",
    )

    description: Optional[str] = Field(
        title="Check Description",
        description="An optional description for the health check (not parsed)",
        default=None,
    )

    args: DNSArgs | FileArgs | HTTPArgs | ICMPArgs | NTPArgs | ShellArgs | TCPArgs = (
        Field(
            title="Check Arguments",
            description="The arguments for the health check check",
            discriminator="method",
        )
    )

    path_id: Optional[
        Annotated[NonNegativeInt, Field(le=4294967295, strict=True)] | IPv4Address
    ] = Field(
        title="Path ID",
        description="The path ID to announce with the route",
        default=None,
    )

    prefixes: list[IPvAnyNetwork] = Field(
        title="Prefixes",
        description="The list of IP addresses or prefixes that will be advertised",
    )

    nexthop: IPvAnyAddress | Literal["self"] = Field(
        title="Next Hop",
        description="The next-hop address for the prefix that will be advertised",
    )

    metric: Optional[NonNegativeInt] = Field(
        title="Metric",
        description="The metric to advertise with the route",
        le=4294967295,
        default=None,
    )

    metric_down: Optional[NonNegativeInt] = Field(
        title="Down Metric",
        description="If the health check fails, continue to announce the route but with this metric",
        le=4294967295,
        default=None,
    )

    local_preference: Optional[NonNegativeInt] = Field(
        title="Local Preference",
        description="The local preference to announce with the route",
        le=4294967295,
        default=None,
    )

    disable: Optional[Path] = Field(
        title="Disable File",
        description="If the file exists, the route will be withdrawn and the check will be disabled",
        default=None,
    )

    neighbors: Optional[list[IPvAnyAddress]] = Field(
        title="Neighbors",
        description="Filter the route to only be announced to the specified neighbors",
        default=None,
    )

    communities: Optional[list[str]] = Field(
        title="BGP Communities",
        description="Announce the route with the specified BGP communities",
        default=None,
    )

    as_path: Optional[
        Annotated[PositiveInt, Field(le=4294967294, strict=True)] | str
    ] = Field(
        title="AS Path",
        description="Announce the route with the specified AS path",
        default=None,
    )

    interval: PositiveInt | PositiveFloat = Field(
        title="Check Interval",
        description="The interval in seconds between checks",
        default=15,
    )

    rise: NonNegativeInt = Field(
        title="Rise",
        description="The number of consecutive successful checks required to mark the check as healthy",
        default=3,
    )

    fall: NonNegativeInt = Field(
        title="Fall",
        description="The number of consecutive failed checks required to mark the check as failed",
        default=3,
    )

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_prefixes_list(cls, values: dict) -> dict:
        """
        Set the list of prefixes to advertise to a list if it is not already
        """
        # Ensure that the prefixes are defined
        if "prefixes" not in values or values["prefixes"] is None:
            raise ValueError(
                f"One or more prefixes to advertise must be defined for check {values['name']}"
            )

        # Ensure that the prefixes are a list
        if not isinstance(values["prefixes"], list):
            # Create list of prefixes
            prefixes: list[str] = [values["prefixes"]]
            # Set the prefixes to the list generated
            values["prefixes"] = prefixes

        # Return the correctly set values
        return values

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_path_id(cls, values: dict) -> dict:
        """
        If the path ID attribute is defined and it is a string, convert to an IPv4Address object
        """
        # Skip processing if not defined
        if "path_id" not in values or values["path_id"] is None:
            return values

        # Check if the path ID is a string
        if isinstance(values["path_id"], str):
            try:
                values["path_id"] = IPv4Address(values["path_id"])
            except Exception as exc:
                raise ValueError(
                    f"Invalid path ID '{values['path_id']}' for check {values['name']} defined: "
                    "It must be an integer in the range 0-4294967295 or an IPv4 address"
                ) from exc

        # Return the path ID
        return values

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def validate_communities_string(cls, values: dict) -> dict:
        """
        If the communities attribute is set, ensure that it is a string or list of strings
        """
        # Skip processing if not defined
        if "communities" not in values or values["communities"] is None:
            return values

        # Check if the communities is not a list; if not then convert to a list
        if not isinstance(values["communities"], list):
            # Convert to a list
            values["communities"] = [values["communities"]]

        # Loop over each community
        for community in values["communities"]:
            # Raise error if the community is not a string
            if not isinstance(community, str):
                raise ValueError(
                    f"BGP community {community} for check {values['name']} must be a string; current type is "
                    f"{type(community).__name__}"
                )

        # Return the AS path
        return values

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_as_path(cls, values: dict) -> dict:
        """
        If the AS path attribute is set, convert to a string as it may be an integer
        """
        # Skip processing if not defined
        if "as_path" not in values or values["as_path"] is None:
            return values

        # Check if the AS path is an integer
        if isinstance(values["as_path"], int):
            # Convert to a string
            values["as_path"] = f"{values['as_path']}"

        # Return the AS path
        return values

    @field_validator("as_path")
    # pylint: disable=no-self-argument
    def validate_as_path(cls, as_path: str, values: ValidationInfo) -> str:
        """
        Validate the AS path attribute
        """
        # Loop over each AS in the list
        for asn in as_path.split():
            # Convert the ASN to an integer
            asn_int = int(asn)
            # Make sure the AS is within range
            if asn_int < 1 or asn_int > 4294967295:
                raise ValueError(
                    f"Invalid AS number '{asn}' defined in the AS path for check {values.data['name']}: "
                    "It must be an integer in the range 1-4294967295"
                )

        # Return the AS path
        return as_path

    @field_validator("communities")
    # pylint: disable=no-self-argument
    def validate_communities(cls, communities: list[str]) -> list[str]:
        """
        Validate BGP communities
        """
        # Create community regex
        community_regex = re.compile(
            r"^(\d+:\d+(:\d+)?|no-export|no-advertise|no-export-subconfed|internet|local-as)$"
        )

        # Loop over each community
        for community in communities:
            # Validate the community matches pattern
            if not community_regex.match(community):
                raise ValueError(f"BGP community {community} is not valid")

        # Return the communities
        return communities

    @field_validator("nexthop")
    # pylint: disable=no-self-argument
    def validate_nexthop(
        cls,
        nexthop: IPv4Address | IPv6Address | Literal["self"],
        values: ValidationInfo,
    ) -> IPv4Address | IPv6Address | Literal["self"]:
        """
        Validate the next hop attribute matches the prefix address families
        """
        # Check if the next hop attribute is self
        if nexthop == "self":
            # Return without further validation as the prefixes cannot be validated
            return nexthop

        # Check if the next hop is IPv4 and any prefixes are not IPv4
        if isinstance(nexthop, IPv4Address) and not all(
            isinstance(prefix, IPv4Network) for prefix in values.data["prefixes"]
        ):
            raise ValueError(
                f"Next hop address '{nexthop}' defined for check {values.data['name']} must match the address family "
                "of the prefixes that are to be advertised (the prefixes must be IPv4 only if the next hop is IPv4)"
            )

        # Check if the next hop is IPv6 and any prefixes are not IPv6
        if isinstance(nexthop, IPv6Address) and not all(
            isinstance(prefix, IPv6Network) for prefix in values.data["prefixes"]
        ):
            raise ValueError(
                f"Next hop address '{nexthop}' defined for check {values.data['name']} must match the address family "
                "of the prefixes that are to be advertised (the prefixes must be IPv6 only if the next hop is IPv6)"
            )

        # Return the next hop
        return nexthop



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Check Method - HTTP
"""

import warnings
from pprint import pformat
from ipaddress import ip_address, IPv6Address

import httpx

from ..checkresult import CheckResult
from ..settings.checkargs.httpargs import HTTPArgs
from ._remote import Remote


class HTTP(Remote):
    """
    Run a HTTP health check
    """

    def check_one(self, addr: str) -> CheckResult:
        """
        Run the HTTP health check
        """
        # Set type for MyPy
        self.args: HTTPArgs

        # Send the HTTP request and get response
        try:
            response = self._request(addr=addr)
        except httpx.ConnectError as exc:
            return CheckResult(
                success=False,
                message="HTTP request connection error",
                error=f"Connection error to {addr}: {exc}",
            )
        except Exception as exc:
            return CheckResult(success=False, message=f"HTTP request failed: {exc}")

        # Perform validation on the response and get the check result object
        result = self._validate(response=response)

        # Return the check result
        return result

    def _validate(self, response: httpx.Response) -> CheckResult:
        """
        Validate the HTTP response received
        """
        # Check if a specific response code is required; if so validate it matches
        if (
            self.args.expected_status
            and response.status_code not in self.args.expected_status
        ):
            return CheckResult(
                success=False,
                message=f"HTTP status code {response.status_code} does not match expected status code(s)",
                error=(
                    f"{response.status_code} not listed in the status codes: "
                    f"{', '.join(map(str, self.args.expected_status))}"
                ),
            )

        # Check if a successful response code is required; if so validate it matches
        if self.args.require_status and response.status_code != httpx.codes.OK:
            return CheckResult(
                success=False,
                message=f"HTTP status code {response.status_code} does not match expected status code(s)",
                error=f"{response.status_code} could not be confirmed as a valid status code",
            )

        # Check if a response pattern is required; if so validate it matches
        if self.args.response and not self.args.response.search(response.text):
            return CheckResult(
                success=False,
                message="Response pattern not found in response body",
                error=f"The response pattern {self.args.response.pattern!r} did not match the response body",
            )

        # Return a valid check result
        return CheckResult(
            success=True,
            message="HTTP response was valid and passed all checks",
            output=f"HTTP response received in {response.elapsed.total_seconds():.3f} seconds",
        )

    def _request(self, addr: str) -> httpx.Response:
        """
        Send the HTTP request and return the response
        """
        # Get the URL to make request to
        url = self._create_url(addr=addr)

        # Create the headers for the request
        headers = self._create_headers()

        # Set up authentication if required
        authentication = self._create_auth()

        # Get any extensions required
        extensions = self._create_extensions()

        # Send the request of the appropriate type
        with httpx.Client(
            headers=headers,
            verify=self.args.verify_ssl,
            http2=self.args.http2,
            timeout=self.args.http_timeout,
            auth=authentication,
        ) as client:
            # If the request is a GET request, no data can be included
            if self.args.request_method.lower() == "get":
                response = client.get(
                    url=url,
                    params=self.args.url.query,
                    extensions=extensions,
                )
            else:
                response = getattr(client, self.args.request_method.lower())(
                    url=url,
                    params=self.args.url.query,
                    data=self.args.data,
                    extensions=extensions,
                )

        # Return the response
        return response

    def _create_headers(self) -> dict[str, str]:
        """
        Create the HTTP headers to use for the HTTP request
        """
        self.log.bind(event="debug").debug("Creating HTTP headers")

        # Create headers dict
        headers: dict[str, str] = {}

        # Set user agent header
        headers["User-Agent"] = self.args.user_agent

        # If the address to check is not an IP address set the host header
        if self.args.url.host and not self._is_ip(self.args.url.host):
            # Host isn't an IP; set host header
            headers["Host"] = self.args.url.host

        # Add any extra headers if required
        # As the values may be an integer/float, coerce them into a string
        if self.args.headers:
            for k, val in self.args.headers.items():
                headers[k] = f"{val}"

        # Log the headers that will be used
        self.log.bind(event="datadump").trace(
            "HTTP headers created for client:\n{headers}",
            headers=pformat(headers, indent=4, width=120),
        )

        # Return the headers
        return headers

    def _create_url(self, addr: str) -> str:
        """
        Build the URL that the HTTP health check should be made to
        """
        self.log.bind(event="debug").debug("Creating HTTP URL")

        # Create URL template
        template = "{scheme}://{host}{port}{path}"

        # Set port if defined
        if self.args.url.port:
            port = f":{self.args.url.port}"
        else:
            port = ""

        # Set path if defined
        if self.args.url.path:
            path = self.args.url.path
        else:
            path = "/"

        # If the IP address to test is an IPv6 IP, enclose it with square brackets
        if isinstance(ip_address(addr), IPv6Address):
            addr = f"[{addr}]"

        # Generate the URL
        url = template.format(
            scheme=self.args.url.scheme,
            host=addr,
            port=port,
            path=path,
        )

        # Log the URL that will be used
        self.log.bind(event="debug").debug("HTTP URL created: {url}", url=url)

        # Return the URL
        return url

    def _create_auth(self) -> tuple[str, str] | None:
        """
        Set up a HTTP basic authentication object if required
        """
        # Only proceed if the username and password are set
        if self.args.url.username and self.args.url.password:
            # Return tuple of username/password
            return (self.args.url.username, self.args.url.password)

        # No auth configured
        return None

    def _create_extensions(self) -> dict[str, str] | None:
        """
        Set up any extensions required for the HTTP request

        Currently this only configures SNI if required but this may be extended in future
        """
        # If the scheme is not HTTPS, skip
        if self.args.url.scheme != "https":
            return None

        # If there is no host available, skip
        if not self.args.url.host:
            return None

        # If the host is an IP address, skip
        if self._is_ip(self.args.url.host):
            return None

        # Create empty dict of extensions
        extensions: dict[str, str] = {}

        # Set up the SNI extension
        extensions["sni"] = self.args.url.host

        # Return the extensions
        return extensions

    # Check if the supplied value is an IP address
    @staticmethod
    def _is_ip(value: str) -> bool:
        """
        Check if the supplied value is an IP address
        """
        try:
            ip_address(value)
            return True
        except ValueError:
            return False



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Store configuration for Sentry SDK
"""

from pydantic import Field, HttpUrl, NonNegativeFloat

from ._base import Base


class Sentry(Base):
    """
    Configuration for sentry SDK
    """

    dsn: HttpUrl = Field(
        title="DSN",
        description="The DSN to send to send Sentry logs to",
    )

    enabled: bool = Field(
        title="Enable",
        description="If Sentry should be sending events to the specified DSN",
        default=True,
    )

    sample_rate: NonNegativeFloat = Field(
        title="Sample Rate",
        description="The rate at which to sample events for Sentry",
        default=1.0,
        ge=0.0,
    )

    profiles_sample_rate: NonNegativeFloat = Field(
        title="Profile Sample Rate",
        description="If set, the Sentry profiling feature will be enabled at the specified sampling rate",
        default=0.0,
        ge=0.0,
    )

    attach_stacktrace: bool = Field(
        title="Attach Stacktrace",
        description="If Sentry should attach a stacktrace to all messages logged (not only exceptions)",
        default=False,
    )

    include_local_variables: bool = Field(
        title="Include Local Variables",
        description="If Sentry should capture and send local variables to send with events",
        default=True,
    )

    debug: bool = Field(
        title="Debug",
        description="Enable debugging mode for Sentry",
        default=False,
    )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Store the current state of a health check instance
"""

from datetime import datetime
from pprint import pformat
from typing import Literal, Optional

from pydantic import ConfigDict, BaseModel, Field, PositiveInt

from .checkresult import CheckResult


class CheckState(BaseModel):
    """
    Store the current state information for a health check
    """

    state: Literal["up", "down", "rising", "falling", "disabled", "startup"] = Field(
        title="Check State",
        description="The current state of the health check",
    )

    advertised: bool = Field(
        title="Advertised",
        description="Whether the prefixes are currently being advertised",
    )

    last_result: Optional[CheckResult] = Field(
        title="Last Check Result",
        description="The CheckResult object from the last health check",
        default=None,
    )

    down_since: Optional[datetime] = Field(
        title="Down Since",
        description="The time the health check entered the down state",
        default=None,
    )

    up_since: Optional[datetime] = Field(
        title="Up Since",
        description="The time the health check entered the up state",
        default=None,
    )

    rise: Optional[PositiveInt] = Field(
        title="Rise Count",
        description="If the service is not in an 'up' state, the number of consecutive successful health checks",
        default=None,
    )

    fall: Optional[PositiveInt] = Field(
        title="Fall Count",
        description="If the service is not in a 'down' state, the number of consecutive failed health checks",
        default=None,
    )
    model_config = ConfigDict(extra="forbid", frozen=True)

    @property
    def pretty(self) -> str:
        """
        Return a pretty printed version of the current health check state
        """
        return pformat(self.model_dump(), indent=4, width=120)



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Store configuration for logging to a file and/or syslog
"""

from __future__ import annotations

import re
import os
import stat
from pathlib import Path
from ipaddress import ip_address
from typing import Literal

from pydantic import Field, field_validator, PositiveInt, FilePath

from ._base import Base


class Syslog(Base):
    """
    Define and store parsed configuration for syslog based logging
    """

    method: Literal["syslog"] = Field(
        title="Log Method",
        description="Enable logging to a syslog server hostname, IP address or socket",
    )

    structured: bool = Field(
        title="Structured Logging",
        description="Use structured logging format",
        default=True,
    )

    destination: Literal["/dev/log"] | FilePath | str = Field(
        title="Syslog Destination",
        description="The destination IP address or hostname to send messages to",
        default="/dev/log",
    )

    port: PositiveInt = Field(
        title="Syslog Port",
        description="The port to send syslog messages to",
        default=514,
        le=65535,
    )

    protocol: Literal["tcp", "udp"] = Field(
        title="Syslog Protocol",
        description="The protocol to send syslog messages with",
        default="udp",
    )

    @field_validator("destination")
    # pylint: disable=no-self-argument
    def validate_destination(
        cls, destination: Literal["/dev/log"] | str
    ) -> Literal["/dev/log"] | str:
        """
        Perform very basic validation for the destination address to send syslog messages
        """
        # Skip validation for "/dev/log"
        if destination == "/dev/log":
            return destination

        # Check if destination is an IPv4 or IPv6 address
        try:
            ip_address(destination)
        except ValueError:
            # Test if it is a path to a socket
            if os.path.exists(destination):
                # Path exists, make sure it is a socket
                mode = os.stat(destination).st_mode
                if not stat.S_ISSOCK(mode):
                    raise ValueError(
                        f"Logging destination {destination} is not a socket"
                    )

                # Convert the destination into a file path
                destination = Path(destination).resolve()
                assert isinstance(destination, Path)

                # Make sure the socket is writable
                if not os.access(destination, os.W_OK):
                    raise ValueError(
                        f"Destination logging socket {destination} is not writable"
                    )

            # Test if it is a hostname
            else:
                host_regex = re.compile(
                    r"^[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9](\.[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])*$"
                )
                if not host_regex.match(destination):
                    raise ValueError(  # pylint: disable=raise-missing-from
                        f"Invalid destination address '{destination}' for syslog messages"
                    )

        # Return as it seems to be valid enough to use
        return destination



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

TCP health check arguments
"""

from typing import Literal

from pydantic import Field, field_validator, PositiveInt, PositiveFloat
from pydantic_core.core_schema import ValidationInfo

from ._remote import Remote


class TCPArgs(Remote):
    """
    Model for TCP check arguments
    """

    method: Literal["tcp"] = Field(
        title="TCO Check Method",
        description="Open a TCP connection to a host/port and ensure it is successful",
    )

    port: PositiveInt = Field(
        title="Port",
        description="The port to attempt connecting to",
        le=65535,
    )

    tcp_timeout: PositiveInt | PositiveFloat = Field(
        title="TCP Timeout",
        description="The timeout for the TCP connection",
        default=5,
    )

    @field_validator("tcp_timeout")
    # pylint: disable=no-self-argument
    def validate_tcp_timeout(
        cls, tcp_timeout: int | float, values: ValidationInfo
    ) -> int | float:
        """
        Ensure that the TCP timeout is a valid value
        """
        # Check if the TCP timeout is higher than the check timeout
        if tcp_timeout > values.data["timeout"]:
            raise ValueError(
                f"Invalid TCP timeout '{tcp_timeout}' defined for check: "
                f"It must be less than the check timeout of {values.data['timeout']} seconds"
            )

        # Return the TCP timeout
        return tcp_timeout



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

File health check arguments
"""

from os import W_OK, access
from pathlib import Path
from typing import Literal

from pydantic import Field, field_validator

from ._base import Base


class FileArgs(Base):
    """
    Model for file check arguments
    """

    method: Literal["file"] = Field(
        title="File Check Method",
        description="Test if a file exists or doesn't exist",
    )

    path: Path = Field(
        title="File Path",
        description="The path to the file that should be checked",
    )

    exists: bool = Field(
        title="File Exists",
        description="If set to false the check will fail if the file exists",
        default=True,
    )

    @field_validator("path")
    # pylint: disable=no-self-argument
    def validate_path(cls, path: Path) -> Path:
        """Validate that the file parent directory can be read

        Args:
            path (Path): The path to the file

        Returns:
            Path: The original path
        """
        # Try to access the parent directory
        if not access(path.parent, W_OK):
            raise ValueError(
                f"File path '{path}' parent directory '{path.parent}' is not readable; the health check cannot work"
            )

        # Return the path
        return path



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Check Method - Shell Command
"""

from __future__ import annotations

import subprocess
from typing import Tuple

from ..checkresult import CheckResult
from ..settings.checkargs.shellargs import ShellArgs
from ._base import Base


# pylint: disable=too-few-public-methods
class Shell(Base):
    """
    Run a shell command health check
    """

    @staticmethod
    def _to_str(
        stdout: bytes | None, stderr: bytes | None
    ) -> Tuple[str | None, str | None]:
        """
        Convert stdout and stderr to strings if defined
        """
        # If there is stdout convert to string
        stdout_str: str | None = bytes.decode(stdout, "utf-8") if stdout else None

        # If there is stderr convert to string
        stderr_str: str | None = bytes.decode(stderr, "utf-8") if stderr else None

        return stdout_str, stderr_str

    def check(self) -> CheckResult:  # NOSONAR
        """
        Run the health check
        """
        # Set type for MyPy
        self.args: ShellArgs

        # Execute the command
        try:
            result = subprocess.run(  # nosemgrep:python.lang.security.audit.subprocess-shell-true.subprocess-shell-true
                self.args.command,
                check=True,
                shell=True,
                env=self.args.environment,
            )
        except subprocess.CalledProcessError as error:
            # Convert stdout/stderr to strings if defined
            stdout, stderr = self._to_str(error.stdout, error.stderr)
            # Return the health check result
            return CheckResult(
                success=False,
                message=f"Command returned non-zero exit code: {error.returncode}",
                output=stdout,
                error=stderr,
            )
        except Exception as exc:  # pylint: disable=broad-except
            return CheckResult(
                success=False,
                message="Shell command failed to execute",
                error=f"Exception raised: {exc}",
                exception=exc,
            )

        # Convert stdout/stderr to strings if defined
        stdout, stderr = self._to_str(result.stdout, result.stderr)

        # Return the result
        return CheckResult(
            success=True,
            message="Shell command executed successfully",
            output=stdout,
            error=stderr,
        )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Manage the announcement and withdrawal of routes to ExaBGP
"""

from __future__ import annotations

import re
from sys import stdout
from pprint import pformat
from typing import Literal

import loguru

from .settings.check import Check


class Announcer:
    """
    Announce and withdraw routes from ExaBGP
    """

    def __init__(self, check: Check, log_context: loguru.Logger):
        """Set up the new ExaCheck announcer object

        Args:
            check (Check): The health check configuration object
            log_context (loguru.Logger): The loguru logging context
        """

        # Set the logging context
        self.log = log_context.bind(subsystem="announcer")

        # Build the list of routes to manage
        self.routes = self.generate_routes(check=check)

        # Log completion of setup
        self.log.bind(event="debug").info("New route announcer object setup")

    def announce(self, metric: int | None = None) -> None:
        """Announce a route to ExaBGP

        Args:
            metric (int | None, optional): The metric for the route. Defaults to None.
        """
        # Log the route advertisement event
        self.log.bind(event="info").info("Advertising routes")

        # Advertise the routes
        self.send_command(command="announce", metric=metric)

    def withdraw(self, metric: int | None = None, silent: bool = False) -> None:
        """Withdraw a route from ExaBGP

        Args:
            metric (int | None, optional): The metric for the route. Defaults to None.
        """
        # Log the route withdraw event
        if not silent:
            self.log.bind(event="info").info("Withdrawing routes")

        # Withdraw the routes
        self.send_command(command="withdraw", metric=metric, silent=silent)

    def send_command(
        self,
        command: Literal["announce", "withdraw"],
        metric: int | None = None,
        silent: bool = False,
    ) -> None:
        """Send the announce or withdraw command to ExaBGP

        The command is sent to ExaBGP via stdout.

        Args:
            command (Literal["announce","withdraw]): If the route needs to be withdrawn or announced/advertised.
            metric (int | None, optional): The metric for the route. Defaults to None.
        """
        # Test if a metric was defined; if set, create the metric string
        if metric:
            med = f"med {metric}"
        else:
            # No metric defined, just set to empty variable
            med = None

        # Loop over each route
        for route in self.routes:
            # Format the route with command and MED/metric
            route_string = route.format(command=command, metric=med)

            # Log the command being sent
            if not silent:
                self.log.bind(event="datadump").trace(
                    "Sending command to ExaBGP:\n{route}",
                    route=route_string,
                )

            # Send the command to ExaBGP
            try:
                print(route_string)
                stdout.flush()
            except Exception as exc:
                # Log the error sending the command
                if not silent:
                    self.log.bind(event="error").error(
                        "Error sending command to ExaBGP: {exc}",
                        exc=exc,
                    )

    def generate_routes(self, check: Check) -> list[str]:
        """Generate the route commands used by ExaBGP to announce/withdraw routes

        Args:
            check (Check): The health check configuration object.

        Returns:
            list[str]: The list of formatted routes as strings ready for ExaBGP.
        """
        # Send log about generation
        self.log.bind(event="info").debug("Generating route(s) to announce/withdraw")

        # Create empty list to store the route to use as a template
        route_template: list[str] = []

        # Check if there are neighbors
        if check.neighbors:
            # Create a list of neighbors
            neighbors = [f"neighbor {neighbor}" for neighbor in check.neighbors]
            # Add the neighbors to route
            route_template.append(", ".join(neighbors))

        # Add the announce/withdraw command which will be formatted when actually announcing/withdrawing
        route_template.append("{{command}}")

        # Add the route prefix
        route_template.append("route {prefix}")

        # Add the next hop
        route_template.append(f"next-hop {check.nexthop}")

        # Add the metric placeholder if required
        if check.metric or check.metric_down:
            route_template.append("{{metric}}")

        # Add the local preference if defined
        if check.local_preference:
            route_template.append(f"local-preference {check.local_preference}")

        # Add BGP communities if defined
        if check.communities:
            route_template.append(self._generate_communities(check.communities))

        # Add the AS path if defined
        if check.as_path:
            route_template.append(f"as-path {check.as_path}")

        # Add the path ID if required
        if check.path_id:
            route_template.append(f"path-information {check.path_id}")

        # Join route template
        route_template_string = " ".join(route_template)

        # Log the generated route template
        self.log.bind(event="info").debug(
            "Generated route template: {route_template_string}",
            route_template_string=route_template_string,
        )

        # Create empty list to store the routes
        routes: list[str] = []

        # Loop over each prefix defined in the check
        for prefix in check.prefixes:
            # Apply prefix to route template and append to routes
            routes.append(route_template_string.format(prefix=prefix))

        # Log the completed routes
        self.log.bind(event="datadump").trace(
            "Generated routes from template:\n{routes}",
            routes=pformat(routes, indent=4, width=120),
        )

        # Return routes
        return routes

    def _generate_communities(self, communities: list[str]) -> str:
        """Generate a list of communities to insert into the route string to advertise

        This is required as the community type needs to be determined (to see if it is a large or extended community).

        Args:
            communities (list[str]): The list of BGP communities to insert into the route string

        Returns:
            str: The formatted communities as a single string.
        """
        # Log community generation
        self.log.bind(event="debug").debug(
            "Generating communities to insert into route"
        )
        self.log.bind(event="datadump").trace(
            "Raw communities:\n{communities}",
            communities=pformat(communities, indent=4, width=120),
        )

        # Create community regex patterns
        standard_pattern = re.compile(r"^\d{1,10}:\d{1,10}$")
        large_pattern = re.compile(r"^\d{1,10}:\d{1,10}:\d{1,10}$")

        # Create empty list to store each form of communities
        standard: list[str] = []
        large: list[str] = []
        extended: list[str] = []

        # Loop over each community provided
        for community in communities:
            # Check if the community is a standard community
            if standard_pattern.match(community):
                # Add the community to the standard list
                standard.append(community)
                continue

            # Check if the community is a large community
            if large_pattern.match(community):
                # Add the community to the large list
                large.append(community)
                continue

            # Make the assumption the community is extended if reaching this point
            extended.append(community)

        # Create list to store the communities after formatting
        community_string: list[str] = []

        # Check if there are any standard communities
        if standard:
            # Add the standard communities to the route
            community_string.append(f"community [{' '.join(standard)}]")

        # Check if there are any standard communities
        if large:
            # Add the standard communities to the route
            community_string.append(f"large-community [{' '.join(large)}]")

        # Check if there are any standard communities
        if extended:
            # Add the standard communities to the route
            community_string.append(f"extended-community [{' '.join(extended)}]")

        # Return the communities joined together
        communities_joined = " ".join(community_string)
        self.log.bind(event="datadump").trace(
            "Generated communities string:\n{community_string}",
            community_string=communities_joined,
        )
        return communities_joined



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Base Check Method
"""

from __future__ import annotations

from abc import ABC, abstractmethod

import loguru

from ..settings.checkargs import CheckArgs
from ..checkresult import CheckResult


# pylint: disable=too-few-public-methods
class Base(ABC):
    """
    The base health check class
    """

    def __init__(
        self,
        log_context: loguru.Logger,
        args: CheckArgs,
    ):
        """
        Initialize the class
        """
        # Define log context
        self.log = log_context.bind(subsystem="healthcheck")

        # Define check args
        self.args = args

        # Complete setup
        self.log.bind(event="debug").info("Health check setup complete")

    @abstractmethod
    def check(self) -> CheckResult:
        """Function that is called each healthcheck interval.

        Your health check must return a "CheckResult" object with the results of the health check.
        Any exceptions should be handled in this function with a textual representation of the exception stored
        in the CheckResult object.

        Returns:
            CheckResult: The results of the health check
        """

    def _log_result(self, result: CheckResult) -> None:
        """
        Log a check result
        """
        # Log the data
        self.log.bind(event="debug").debug(
            "Health check returned {status}",
            status="success" if result.success else "failure",
        )
        self.log.bind(event="datadump").trace(
            "Health check result:\n{result}",
            result=result.pretty,
        )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Shell command health check arguments
"""

from typing import Literal, Optional

from pydantic import Field

from ._base import Base


class ShellArgs(Base):
    """
    Model for shell command check arguments
    """

    method: Literal["shell"] = Field(
        title="Shell Check Method",
        description="Run a script to determine if a service is healthy or not based on the exit code",
    )

    command: str = Field(
        title="Shell Command",
        description="The shell command to run for the health check",
    )

    environment: Optional[dict[str, str]] = Field(
        title="Environment Variables",
        description="Environment variables to set when running the command",
        default=None,
    )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Health check work class - handles the tasks of running the health check
and announcing or withdrawing routes as needed.
"""

from __future__ import annotations

from datetime import datetime
from signal import signal, SIGTERM, SIGINT
import sys

from loguru import logger

from . import methods
from .announcer import Announcer
from .checkexecutor import CheckExecutor
from .checkresult import CheckResult
from .checkstate import CheckState
from .procname import ProcName
from .settings.check import Check
from .sleeper import Sleeper
from .notifications import Notifications


# pylint: disable=too-few-public-methods,too-many-instance-attributes
class Worker:
    """
    The ExaCheck worker process for a health check; the worker will call the checks and handle
    route advertisement/withdrawal as needed
    """

    def __init__(self, check: Check, notifications: Notifications):
        """
        Set up the new ExaCheck worker object
        """
        # Bind logging process name
        self.log = logger.bind(check_name=check.name, subsystem="worker")

        # Set the check data
        self.check: Check = check

        # Set the notifications object
        self.notifications = notifications

        # Create process name manager object
        self.procname = ProcName(
            base=f"ExaCheck Worker [{self.check.name}]",
            log_context=self.log,
        )
        # Set status to "starting"
        self.procname.update(message="Startup")

        # Create announcer
        self.announcer = Announcer(check=self.check, log_context=self.log)

        # Set the initial check state
        self.check_state = CheckState(
            state="startup",
            advertised=False,
        )

        # Set check method
        self.check_method = check.args.method

        # Setup done
        self.log.bind(event="info").info(
            "ExaCheck worker object created; beginning checks"
        )

        # Start the check loop
        self.run()

    def run(self) -> None:
        """
        Run the health check process loop
        """
        # Create the check executor
        executor = CheckExecutor(method=self.method, log_context=self.log)

        # Create signal handler to withdraw routes on SIGTERM/SIGINT
        signal(SIGINT, self.cleanup)
        signal(SIGTERM, self.cleanup)

        # Start infinite loop
        while True:
            # Log start of health check
            self.log.bind(event="debug").debug("Health check loop start")

            # Create the sleeper object and start timer
            sleeper = Sleeper(
                interval=self.check.interval,
                log_context=self.log,
            )

            # Update process name
            self.procname.update(
                message="Performing health check",
                status=self._procname_status,
            )

            # Test if the check has a disable file and if that disable file exists
            if self.check.disable and self.check.disable.exists():
                # The disable file exists
                self.log.bind(event="info").info(
                    "Health check is disabled by disable file '{disable_file}'; no health check executing",
                    disable_file=self.check.disable,
                )

                # Create the check result
                result = CheckResult(
                    success=False,
                    message=f"Health check is disabled by disable file '{self.check.disable}'",
                    disabled=True,
                )

                # Call the failure handler
                self.failure(result=result)

            else:
                # Perform the health check
                result = executor.execute()

            # Check the result
            if result.success:
                # Call the success handler
                self.success(result=result)
            else:
                # Call the failure handler
                self.failure(result=result)

            # Finish the sleep timer
            sleeper.finish()

            # Update process name
            self.procname.update(
                message=f"Sleeping for {sleeper.sleep_time:.3f} seconds",
                status=self._procname_status,
            )

            # Sleep for the remaining time
            sleeper.sleep()

    @property
    def method(self) -> methods.CheckMethods:
        """
        Generate and return the class for the type of health check requested
        """
        # Ensure the check method is available
        if self.check_method not in methods.METHODS:
            self.log.bind(event="error").critical(
                "No health check method '{method}' available",
                method=self.check_method,
            )
            raise NotImplementedError(
                f"Could not get check method: {self.check_method}"
            )

        # Look up the class for the check
        _, cls = methods.METHODS[self.check_method]

        # Set up and return the check
        return cls(
            log_context=self.log,
            args=self.check.args,
        )

    def success(self, result: CheckResult) -> None:
        """
        Called when the health check returns a successful status
        """
        # Get existing check state object
        state = self.check_state

        # Check if the route is currently advertised
        if state.advertised:
            # The route is already advertised, create new check state
            self.log.bind(event="debug").info(
                "Health check successful and prefixes are already advertised"
            )
            self.check_state = CheckState(
                state="up",
                advertised=True,
                last_result=result,
                up_since=state.up_since,
            )
            # Return
            return

        # Check if the service has risen
        if state.rise and (state.rise + 1) >= self.check.rise:
            # Service is now risen, call function to announce route
            self.log.bind(event="announce").success(
                "Health check successful and service has risen; prefixes will be advertised"
            )
            self.announce()
            # Create the new check state
            self.check_state = CheckState(
                state="up",
                advertised=True,
                last_result=result,
                up_since=datetime.now(),
            )
            # Return
            return

        # The service does not meet minimum requirements to rise yet
        self.log.bind(event="info").info(
            "Health check successful but service has not risen yet",
        )

        # If there is an existing rise value, add 1 otherwise set to 1
        if state.rise:
            rise = state.rise + 1
        else:
            rise = 1

        # Create the new check state
        self.check_state = CheckState(
            state="rising",
            advertised=False,
            last_result=result,
            down_since=state.down_since,
            rise=rise,
        )

    def failure(self, result: CheckResult) -> None:
        """
        Called when the health check returns a failed status
        """
        # Get existing check state object
        state = self.check_state

        # Check if the failure is due to the service being disabled
        if result.disabled:
            # Check if the route is currently advertised
            if state.advertised:
                # The state is advertised, withdraw the routes
                self.log.bind(event="withdraw").warning(
                    "Service has been disabled; prefixes will be withdrawn"
                )
                self.withdraw()
                # Set the "since" date to now
                since = datetime.now()
            else:
                # The route is not advertised, nothing to do
                since = state.down_since

            # Create the new check state
            self.check_state = CheckState(
                state="disabled",
                advertised=False,
                last_result=result,
                down_since=since,
            )

            # Finish
            return

        # Check if the route is currently advertised
        if not state.advertised:
            # The route is not advertised, create new check state
            self.log.bind(event="debug").info(
                "Health check failure; prefixes are not advertised"
            )
            self.check_state = CheckState(
                state="down",
                advertised=False,
                last_result=result,
                down_since=state.up_since,
            )

            # Return
            return

        # Check if the service is falling
        if state.fall and (state.fall + 1) >= self.check.fall:
            # Service should now be considered down
            self.log.bind(event="withdraw").warning(
                "Health check failure and service has fallen; prefixes will be withdrawn"
            )
            # Withdraw the route
            self.withdraw()
            # Create the new check state
            self.check_state = CheckState(
                state="down",
                advertised=False,
                last_result=result,
                down_since=datetime.now(),
            )
            # Return
            return

        # The service does not meet minimum requirements to fall yet
        self.log.bind(event="info").info(
            "Health check unsuccessful but the service has not fallen yet",
        )

        # If there is an existing fall value, add 1 otherwise set to 1
        if state.fall:
            fall = state.fall + 1
        else:
            fall = 1

        # Create the new check state
        self.check_state = CheckState(
            state="falling",
            advertised=True,
            last_result=result,
            up_since=state.up_since,
            fall=fall,
        )

    def announce(self) -> None:
        """
        Announce the health check route
        """
        # Log
        self.log.bind(event="debug").debug("Call to advertise route")

        # Advertise the route if not advertised yet
        if self.check_state.advertised:
            # Route advertised, nothing to do
            self.log.bind(event="debug").debug(
                "Health check route is already advertised; no action taken"
            )
        else:
            # Advertise the route
            self.log.bind(event="debug").debug(
                "Health check route is not advertised yet; announcing route"
            )
            self.announcer.announce(metric=self.check.metric)

            # Create the message for notification
            message = [
                "Announcing routes for the health check as the service is marked as up.",
                f"The following prefixes will be advertised with the next hop address `{self.check.nexthop}`:",
            ]
            for prefix in self.check.prefixes:
                message.append(f"\n- `{prefix}`")

            # Send notification
            self.notifications.notify(
                event="announce",
                check=self.check.name,
                log_context=self.log,
                title=f"ExaCheck Event - Route Announcement - {self.check.name}",
                message="\n\n".join(message),
            )

    def withdraw(self) -> None:
        """
        Withdraw the health check route
        """
        # Log
        self.log.bind(event="debug").debug("Call to withdraw route advertisement")

        # Withdraw the route if advertised
        if not self.check_state.advertised:
            # Route not advertised, nothing to do
            self.log.bind(event="debug").debug(
                "Health check route is not advertised; no action taken"
            )
            return

        # Withdraw the route
        self.log.bind(event="debug").debug(
            "Health check route is advertised; withdrawing route"
        )
        self.announcer.withdraw(metric=self.check.metric)

        # Get the reason why the route is being withdrawn (disabled or down)
        if self.check_state.state == "disabled":
            reason = "the service has been disabled."
        else:
            reason = "the service has failed."

        # Create the message for notification
        message = [
            f"Withdrawing routes for the health check as {reason}.",
            f"The following prefixes with the next hop address `{self.check.nexthop}` will be withdrawn:",
        ]
        for prefix in self.check.prefixes:
            message.append(f"\n- `{prefix}`")

        # Send notification
        self.notifications.notify(
            event="withdraw",
            check=self.check.name,
            log_context=self.log,
            title=f"ExaCheck Event - Route Withdrawal - {self.check.name}",
            message="\n\n".join(message),
        )

    def cleanup(self, sig: int, frame: object) -> None:
        """Handle clean up of the worker process when the process is terminated

        Args:
            sig (int): The signal that was received
            frame (object): The frame being executed when the signal was received
        """
        # Withdraw all routes if currently advertised
        if self.check_state.advertised:
            self.announcer.withdraw(metric=self.check.metric, silent=True)

        # Finish
        sys.exit(0)

    @property
    def _procname_status(self) -> str:
        """Generate the process name status string"""
        match self.check_state.state:
            case "rising":
                return f"rising ({self.check_state.rise}/{self.check.rise})"
            case "falling":
                return f"falling ({self.check_state.fall}/{self.check.fall})"
            case _:
                return self.check_state.state



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Store internal configuration for ExaCheck
"""

from pydantic import (
    Field,
    PositiveInt,
    PositiveFloat,
)

from ._base import Base


class ExaCheck(Base):
    """
    Configuration for ExaCheck
    """

    live_reload: bool = Field(
        title="Live Reload",
        description="Automatically reload the configuration on any changes",
        default=False,
    )

    monitoring_interval: PositiveInt | PositiveFloat = Field(
        title="Monitoring Interval",
        description=(
            "The interval in seconds between checking the health of each check process/thread and looking for changes "
            "to the configuration file"
        ),
        default=30,
    )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Address family error; the IP address family does not match the requested address family
"""


class AddressFamilyError(Exception):
    """
    Exception raised when a hostname is resolved to an IP address that does not match the requested address family
    """

    def __init__(self, message: str):
        """
        Set up the new AddressFamilyError object
        """
        self.message = message

    def __str__(self):
        return self.message



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

DNS resolution error; host name could not be resolved
"""


class DNSResolutionError(Exception):
    """
    Exception raised when a hostname could not be resolved to an IP address
    """

    def __init__(self, message: str):
        """
        Set up the new DNSResolutionError object
        """
        self.message = message

    def __str__(self):
        return self.message



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Main ExaCheck class
"""

from __future__ import annotations

from errno import ECHILD
from multiprocessing import Process
from os import waitpid, WNOHANG, kill
from pathlib import Path
from pprint import pformat
from time import sleep
from typing import Tuple, cast
from signal import signal, SIGTERM, SIGINT, SIGCHLD
import sys
import importlib.metadata

from loguru import logger

from .exceptions.workerprocesserror import WorkerProcessError
from .logmanager import LogManager
from .procname import ProcName
from .settings.check import Check
from .settings.sentry import Sentry
from .notifications import Notifications
from .sleeper import Sleeper
from .worker import Worker
from .configuration import Configuration


# pylint: disable=too-few-public-methods
class ExaCheck:
    """
    ExaCheck main class
    """

    def __init__(
        self,
        file: Path,
        verbosity: int = 0,
        dry_run: bool = False,
    ):
        """
        Set up the new ExaCheck object
        """
        # Create STDERR logger/Log object
        self.logmanager = LogManager(verbosity=verbosity)

        # Set logging context
        self.log = logger.bind(
            check_name="MASTER",
        )

        # Set logging context to set subsystem to master
        self.log = logger.bind(
            check_name="MASTER",
            subsystem="master",
        )

        # Log setup of object
        self.log.bind(event="info").info(
            "Creating new ExaCheck object with configuration file '{file}'", file=file
        )

        # Load configuration
        self.configuration = Configuration(
            log_context=self.log,
            file=file,
        )

        # Check if dry run
        if dry_run:
            # Log configuration is valid
            self.log.bind(event="info").info(
                "Configuration test only; configuration is valid"
            )
            return

        # If Sentry is to be enabled, try to import the Sentry SDK module and enable it
        if (
            self.configuration.settings.sentry
            and self.configuration.settings.sentry.enabled  # pylint: disable=no-member
        ):
            self._sentry_setup()

        # Configure additional logging if required
        if self.configuration.settings.logging:
            self.logmanager.setup(
                loggers=self.configuration.settings.logging,
            )

        # Configure notifications
        self.notifications = Notifications(
            log_context=self.log,
            configuration=self.configuration.settings.notifications,
        )

        # Log finish of setup
        self.log.bind(event="debug").info("ExaCheck object setup complete")

    def run(self) -> None:
        """
        Run ExaCheck
        """
        # Create process name manager
        self.procname = ProcName(  # pylint: disable=attribute-defined-outside-init
            base=f"ExaCheck Master Process [{self.configuration.settings.file}]",
            log_context=self.log,
        )

        # Set process name
        self.procname.update(message="Starting workers")

        # Create signal handler to handle zombie processes
        signal(SIGCHLD, self._reap)

        # Create jobs
        self.jobs = self._start_processes()

        # Notify that ExaCheck is started
        self.notifications.notify(
            event="info",
            title="ExaCheck Startup",
            message="The ExaCheck process has now been started.",
        )

        # Create signal handler to handle process termination and notification
        signal(SIGINT, self.cleanup)
        signal(SIGTERM, self.cleanup)

        # Run loop of doing nothing for now
        self.log.bind(event="info").info(
            "ExaCheck now running; entering monitoring loop"
        )

        # Set new process title
        self.procname.update(message="Sleeping")

        # Perform initial sleep
        self.log.bind(event="debug").debug(
            "Sleeping for {interval} seconds before monitoring loop starts",
            interval=self.configuration.settings.exacheck.monitoring_interval,
        )
        sleep(self.configuration.settings.exacheck.monitoring_interval)

        # Run infinite loop
        while True:
            # Log start of monitoring loop
            self.log.bind(event="info").info("Monitoring loop starting")

            # Create sleep object
            sleeper = Sleeper(
                interval=self.configuration.settings.exacheck.monitoring_interval,
                log_context=self.log,
            )

            # Set new process title
            self.procname.update(message="Monitoring loop start")

            # Loop over each job defined in configuration
            self.log.bind(event="debug").debug(
                "Looping over each worker to ensure it is alive"
            )
            for job in self.jobs:
                # Get the PID of the job
                pid = job[1].pid

                # Make sure the PID exists/is running
                try:
                    kill(pid, 0)
                except Exception:
                    running = False
                else:
                    running = True

                # Ensure that the job is still alive
                if job[1].is_alive() and running:
                    # Worker is alive; log it
                    self.log.bind(event="debug").trace(
                        "Worker '{job_name}' is alive", job_name=job[0].name
                    )

                else:
                    # Worker is not alive; respawn it
                    self.log.bind(event="error").critical(
                        "Worker '{job_name}' is not alive; it will be respawned",
                        job_name=job[0].name,
                    )

                    # Create the new worker process
                    worker = self._create_process(check=job[0])

                    # Start the new worker
                    worker.start()
                    self.log.bind(event="info").info(
                        "Worker '{job_name}' respawned", job_name=job[0].name
                    )

                    # Replace the job with the new one
                    self.jobs[self.jobs.index(job)] = (job[0], worker)

                    # Send notification
                    self.notifications.notify(
                        event="error",
                        title=f"ExaCheck Worker Failure - {job[0].name}",
                        message=f"The ExaCheck worker process for the health check `{job[0].name}` failed and has now been respawned.",
                    )

            # Check if live reloads are enabled
            if self.configuration.settings.exacheck.live_reload:
                # Start configuration file change test
                self.log.bind(event="debug").trace(
                    "Testing if the configuration needs to be reloaded"
                )

                # Check the modification date of the configuration file
                if self.configuration.is_modified():
                    # Log that the configuration file has been modified
                    self.log.bind(event="info").info(
                        "Configuration file has been modified; performing live configuration reload"
                    )
                    # Reload the configuration
                    self._reload_configuration()
                else:
                    # No changes to apply
                    self.log.bind(event="debug").trace(
                        "No configuration changes to apply"
                    )

            else:
                self.log.bind(event="debug").trace(
                    "Live configuration reload disabled; not checking for modifications"
                )

            # Call waitpid in case of zombie processes hanging around
            try:
                waitpid(-1, WNOHANG)
            except Exception:
                pass

            # Finish the sleep timer
            sleeper.finish()

            # Set new process name
            self.procname.update(
                message=f"Sleeping for {sleeper.sleep_time:.3f} seconds"
            )

            # Sleep for the remaining time
            self.log.bind(event="info").debug(
                "Monitoring loop finished; sleeping for {sleep_time:.3f} seconds until next loop",
                sleep_time=sleeper.sleep_time,
            )
            sleeper.sleep()

    def _start_processes(self) -> list[Tuple[Check, Process]]:
        """
        Run ExaCheck using process based health checks
        """
        # Create a list of check processes
        workers: list[Tuple[Check, Process]] = []

        # Start the processes for each check
        for (
            check
        ) in self.configuration.settings.checks:  # pylint: disable=not-an-iterable
            # Log setup
            self.log.bind(event="info").debug(
                "Creating process for check '{check_name}'", check_name=check.name
            )
            self.log.bind(event="datadump").trace(
                "Check '{check_name}' configuration:\n{check_configuration}",
                check_name=check.name,
                check_configuration=pformat(check.model_dump()),
            )

            # Create the process
            worker = self._create_process(check=check)

            # Store the process in the list
            workers.append((check, worker))

        # Loop over each process and start it
        self.log.bind(event="info").info("Starting worker processes")
        for check, worker in workers:
            self.log.bind(event="debug").debug(
                "Starting worker process for '{check_name}'",
                check_name=check.name,
            )
            worker.start()
            self.log.bind(event="info").debug(
                "Worker process for '{check_name}' started",
                check_name=check.name,
            )

        # Return the list of processes
        return workers

    def _create_process(self, check: Check) -> Process:
        """
        Create a single process/worker for a check
        """
        # Create the worker process
        self.log.bind(event="debug").trace(
            "Creating worker process for check {check_name}", check_name=check.name
        )
        try:
            worker = Process(
                target=Worker,
                args=(
                    check,
                    self.notifications,
                ),
                daemon=True,
                name=check.name,
            )
        except Exception as exc:  # pylint: disable=broad-except
            self.log.bind(event="error").exception(
                "Exception creating process for check {check_name}:\n{exc}",
                check_name=check.name,
                exc=exc,
            )
            raise WorkerProcessError from exc

        # Return the created process
        self.log.bind(event="debug").debug(
            "Worker process for check {check_name} created",
            check_name=check.name,
        )
        return worker

    def _sentry_setup(self):
        """
        Set up Sentry error reporting
        """
        # Set type for MyPy
        assert isinstance(self.configuration.settings.sentry, Sentry)
        try:
            import sentry_sdk  # pylint: disable=import-outside-toplevel

            sentry_sdk.init(
                dsn=cast(str, self.configuration.settings.sentry.dsn),
                release=f"exacheck@{importlib.metadata.version('exacheck')}",
                attach_stacktrace=self.configuration.settings.sentry.attach_stacktrace,
                include_local_variables=self.configuration.settings.sentry.include_local_variables,
                debug=self.configuration.settings.sentry.debug,
                traces_sample_rate=self.configuration.settings.sentry.sample_rate,
                profiles_sample_rate=self.configuration.settings.sentry.profiles_sample_rate,
            )
        except ImportError:
            self.log.bind(event="error").error(
                "Sentry SDK not installed; not enabling Sentry reporting"
            )
        except Exception as exc:  # pylint: disable=broad-except
            self.log.bind(event="error").error(
                "Exception enabling Sentry SDK; skipping: {exc}",
                exc=f"{exc}",
            )
        else:
            self.log.bind(event="debug").debug(
                "Enabled Sentry SDK for error reporting to '{dsn}'",
                dsn=self.configuration.settings.sentry.dsn,  # pylint: disable=no-member
            )

    def cleanup(self, sig: int, frame: object) -> None:
        """Handle clean up of the master process when the process is terminated

        Args:
            sig (int): The signal that was received
            frame (object): The frame being executed when the signal was received
        """
        # Send notification
        self.notifications.notify(
            event="info",
            title="ExaCheck Process Terminated",
            message="The ExaCheck process has been terminated.",
        )

        # Exit
        sys.exit(0)

    def _reload_configuration(self) -> None:
        """Perform a reload of the configuration file"""
        # Make a copy of the current configuration object
        current = self.configuration.settings

        # Reload the configuration
        if not self.configuration.reload():
            # Log failure
            self.log.bind(event="error").error(
                "Failed to reload configuration file; skipping reload until next change",
            )
            return

        # Get the new configuration object
        new = self.configuration.settings

        # Check if the notification configuration is different
        if current.notifications != new.notifications:
            # Recreate the notification object
            self.log.bind(event="info").info(
                "Notification configuration has changed; recreating notifications object",
            )
            self.notifications = Notifications(
                log_context=self.log,
                configuration=new.notifications,
            )

        # Loop over all currently defined checks
        for check in current.checks:
            # Test if the check name is not defined anymore
            if not new.get_check_name(check.name):
                # Terminate the worker process
                self._terminate_worker(check)

        # Loop over all newly defined checks
        for check in new.checks:
            # Test if the check is newly defined
            if not current.get_check_name(check.name):
                # Create and start the worker process
                self._start_worker(check)

                # Send notification
                self.notifications.notify(
                    event="info",
                    title=f"ExaCheck Worker Started - {check.name}",
                    message=f"The ExaCheck worker process for the health check `{check.name}` has been started.",
                )

                # Move on to next check
                continue

            # Test if the check has changed
            if new.get_check_name(check.name) != current.get_check_name(check.name):
                # Restart the worker to apply new configuration
                self._restart_worker(check)

    def _restart_worker(self, check: Check) -> None:
        """Restart a worker process"""
        # Log that the check will be restarted
        self.log.bind(event="info").info(
            "Check worker for '{name}' will be restarted",
            name=check.name,
        )

        # Stop the worker process
        self._stop_worker(check)

        # Create the new worker and start process
        self._start_worker(check)

        # Send notification
        self.notifications.notify(
            event="info",
            title=f"ExaCheck Worker Restarted - {check.name}",
            message=f"The ExaCheck worker process for the health check `{check.name}` has been restarted.",
        )

    def _start_worker(self, check: Check) -> None:
        """Create and start a worker process"""
        # Log the startup
        self.log.bind(event="info").info(
            "Creating and starting worker process for '{check_name}'",
            check_name=check.name,
        )

        # Dump the configuration
        self.log.bind(event="datadump").trace(
            "Check '{check_name}' configuration:\n{check_configuration}",
            check_name=check.name,
            check_configuration=pformat(check.model_dump()),
        )

        # Create the new worker process
        worker = self._create_process(check=check)

        # Start worker
        worker.start()

        # Logging
        self.log.bind(event="info").debug(
            "Worker process for '{check_name}' started",
            check_name=check.name,
        )

        # Add to the jobs list
        self.jobs.append((check, worker))

    def _terminate_worker(self, check: Check) -> None:
        """Terminate a worker process"""

        # Log that the check is no longer configured
        self.log.bind(event="info").info(
            "Check worker for '{name}' will be terminated and routes withdrawn",
            name=check.name,
        )

        # Stop the worker
        self._stop_worker(check)

        # Send notification
        self.notifications.notify(
            event="info",
            title=f"ExaCheck Worker Removed - {check.name}",
            message=f"The ExaCheck worker process for the health check `{check.name}` has been removed as it is no longer configured.",
        )

    def _stop_worker(self, check: Check) -> None:
        """Stop a worker process"""
        # Logging
        self.log.bind(event="info").info(
            "Check worker for '{name}' is being stopped",
            name=check.name,
        )

        # Get the worker process associated with the check
        worker = [worker for worker in self.jobs if worker[0].name == check.name][0]

        # Terminate the worker process
        worker[1].terminate()

        # Remove the worker process from the list
        self.jobs.remove(worker)

        # Logging
        self.log.bind(event="info").info(
            "Check worker for '{name}' has been stopped",
            name=check.name,
        )

    def _reap(self, signum, frame) -> None:
        """Handly SIGCHLD signals to reap zombie processes"""
        # Logging
        self.log.bind(event="debug").info("Received SIGCHLD; reaping zombie processes")

        try:
            # Run in loop to catch any processes
            while True:
                # Run waitpid for process
                pid, status = waitpid(-1, WNOHANG)
                # Check if this is the main process
                if pid == 0:
                    # No more processes to reap
                    self.log.bind(event="debug").debug("Finished reaping processes")
                    break

                # Get the exit code of process
                code = status >> 8

                # Log information about it
                self.log.bind(event="debug").info(
                    "Reaped process {pid} with exit code {code}",
                    pid=pid,
                    code=code,
                )

                # Respawn the check
                for check, worker in self.jobs:
                    if worker.pid == pid:
                        self.log.bind(event="info").info(
                            "Worker '{check_name}' has exited; it will be respawned",
                            check_name=check.name,
                        )

                        # Create the new worker process
                        new_worker = self._create_process(check=check)

                        # Start the new worker
                        new_worker.start()

                        # Replace the job with the new one
                        self.jobs[self.jobs.index((check, worker))] = (
                            check,
                            new_worker,
                        )

                        # Send notification
                        self.notifications.notify(
                            event="error",
                            title=f"ExaCheck Worker Failure - {check.name}",
                            message=f"The ExaCheck worker process for the health check `{check.name}` failed and has now been respawned.",
                        )

        except OSError as exc:
            if exc.errno == ECHILD:
                self.log.bind(event="info").info("No child processes to reap")
            else:
                self.log.bind(event="error").error(
                    "Error reaping processes: {exc}",
                    exc=exc,
                )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

HTTP health check arguments
"""

import re
from typing import Literal, Optional, Pattern
from urllib.parse import urlparse

from pydantic import (
    AnyHttpUrl,
    Field,
    model_validator,
    field_validator,
    PositiveInt,
    PositiveFloat,
)
from pydantic_core.core_schema import ValidationInfo

from ...__version__ import __version__
from ._remote import Remote


class HTTPArgs(Remote):
    """
    Model for HTTP check arguments
    """

    method: Literal["http"] = Field(
        title="HTTP Check Method",
        description="Send a HTTP or HTTPS request to a web server and optionally perform various validation",
    )

    host: Optional[str] = Field(  # type: ignore
        title="Hostname/IP Address",
        description="The hostname or IP address the health check will be performed against",
        default=None,
    )

    url: AnyHttpUrl = Field(
        title="URL",
        description="The HTTP/HTTPS URL to check",
    )

    response: Optional[Pattern] = Field(
        title="Response Pattern",
        description="The regular expression to match in the response body of the HTTP request",
        default=None,
    )

    expected_status: Optional[PositiveInt | list[PositiveInt]] = Field(
        title="Expected Status Code",
        description="One or more HTTP status codes that indicate a successful response",
        default=None,
    )

    require_status: bool = Field(
        title="Require 2xx Status Code",
        description="Require a 2xx HTTP status code to indicate a successful response",
        default=True,
    )

    http_timeout: PositiveInt | PositiveFloat = Field(
        title="HTTP Timeout",
        description="The timeout for the HTTP request",
        default=5,
    )

    user_agent: str = Field(
        title="User Agent",
        description="The user agent to send with the HTTP request",
        default=f"ExaCheck HTTP Health Check [v{__version__}]",
    )

    headers: Optional[dict[str, str | int | float]] = Field(
        title="Headers",
        description="Any additional HTTP headers to send with the HTTP request",
        default=None,
    )

    verify_ssl: bool = Field(
        title="Verify SSL",
        description="If making a request to an HTTPS URL, verify the SSL certificate",
        default=False,
    )

    request_method: Literal["GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS"] = Field(
        title="Request Method",
        description="The method for the HTTP request",
        default="GET",
    )

    data: Optional[dict[str, str]] = Field(
        title="Request Data",
        description="The request data to send for POST requests",
        default=None,
    )

    http2: bool = Field(
        title="Enable HTTP/2",
        description="Enable the HTTP/2 protocol for the request",
        default=False,
    )

    @field_validator("http_timeout")
    # pylint: disable=no-self-argument
    def validate_http_timeout(
        cls, http_timeout: int | float, values: ValidationInfo
    ) -> int | float:
        """
        Ensure that the HTTP timeout is a valid value
        """
        # Check if the HTTP timeout is higher than the check timeout
        if http_timeout > values.data["timeout"]:
            raise ValueError(
                f"Invalid HTTP timeout '{http_timeout}' defined for check: "
                f"It must be less than the check timeout of {values.data['timeout']} seconds"
            )

        # Return the HTTP timeout
        return http_timeout

    @field_validator("headers")
    # pylint: disable=no-self-argument
    def validate_header_names(cls, headers: dict[str, str]) -> dict[str, str]:
        """
        Ensure that the header names are valid
        """
        # Make regex to match the valid header name characters
        valid_header_name = re.compile(r"^([\w+\-\=]+)$")

        # Loop over each header and ensure that the header name is valid
        for header_name in headers:
            if not valid_header_name.match(header_name):
                raise ValueError(
                    f"Invalid header name '{header_name}' defined for check: "
                    "Header names must only contain alphanumeric characters and dashes"
                )

        # Return the validated headers
        return headers

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_require_status(cls, values: dict) -> dict:
        """
        If expected_status has a value set require_status must be set to false as the expected status code
        may not be a 2xx code.
        """
        # Test if require_status was set to true specifically
        if "require_status" in values and values["require_status"] is True:
            raise ValueError(
                "The require_status value cannot be set to true when using expected_status"
            )

        # Test if the expected_status value is provided
        if "expected_status" in values and values["expected_status"] is not None:
            # Set require_status false
            values["require_status"] = False

        # Return the new dict
        return values

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_expected_status(cls, values: dict) -> dict:
        """
        If expected_status has been defined and it is an integer, convert it to a list
        """
        # Test if the expected_status value is provided
        if "expected_status" in values and values["expected_status"] is not None:
            # Test if the expected_status value is an integer
            if isinstance(values["expected_status"], int):
                # Convert the expected_status value to a list
                values["expected_status"] = [values["expected_status"]]

        # Return the new dict
        return values

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_host(cls, values: dict) -> dict:
        """
        If no "host" value is provided, set it to the hostname of the URL
        """
        # Only proceed if the URL is provided
        if "url" not in values or values["url"] is None:
            return values

        # Skip if a host value was provided
        if "host" in values and values["host"] is not None:
            return values

        # Attempt to parse the URL
        try:
            url = urlparse(values["url"])
        except ValueError:
            raise ValueError(
                f"Could not parse the supplied URL '{values['url']}' to retrieve hostname; "
                "specify the host parameter manually"
            )

        # Set the host for the URL
        values["host"] = url.hostname

        # Return the values
        return values

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_request_method_case(cls, values: dict) -> dict:
        """
        Set the HTTP method to uppercase
        """
        # Check if a HTTP method was defined by the user
        if "request_method" not in values or values["request_method"] is None:
            # Nothing to do
            return values

        # Set the HTTP method to uppercase
        values["request_method"] = values["request_method"].upper()

        # Return the values
        return values

    @field_validator("data")
    def validate_data(
        cls, data: dict[str, str], values: ValidationInfo
    ) -> dict[str, str]:
        """
        Validate the HTTP request type supports supplying data
        """
        # Make sure the HTTP method is POST
        if values.data["request_method"] != "POST":
            raise ValueError(
                f"Cannot supply data for HTTP request method {values.data['request_method']}"
            )

        # Return the post data
        return data



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

NTP health check arguments
"""

from typing import Literal, Optional

from pydantic import Field, field_validator, PositiveInt, PositiveFloat, NonNegativeInt
from pydantic_core.core_schema import ValidationInfo

from ._remote import Remote


class NTPArgs(Remote):
    """
    Model for NTP check arguments
    """

    method: Literal["ntp"] = Field(
        title="NTP Check Method",
        description="Send a NTP query to a NTP server and validate the response is within the configured thresholds",
    )

    port: PositiveInt = Field(
        title="Port",
        description="The port to send the NTP query to",
        default=123,
        le=65535,
    )

    version: Literal[2, 3] = Field(
        title="NTP Version",
        description="The NTP version to use for the NTP request",
        default=3,
    )

    ntp_timeout: PositiveInt | PositiveFloat = Field(
        title="NTP Timeout",
        description="The timeout for the NTP request",
        default=5,
    )

    max_offset: Optional[PositiveInt | PositiveFloat] = Field(
        title="Maximum NTP Offset",
        description="The maximum offset +/- allowable in seconds before the server is marked as unhealthy",
        default=None,
    )

    max_stratum: Optional[NonNegativeInt] = Field(
        title="Maximum NTP Stratum",
        description="The maximum stratum value before the server is marked as unhealthy",
        default=None,
        le=15,
    )

    @field_validator("ntp_timeout")
    # pylint: disable=no-self-argument
    def validate_ntp_timeout(
        cls, ntp_timeout: int | float, values: ValidationInfo
    ) -> int | float:
        """
        Ensure that the NTP timeout is a valid value
        """
        # Check if the NTP timeout is higher than the check timeout
        if ntp_timeout > values.data["timeout"]:
            raise ValueError(
                f"Invalid NTP timeout '{ntp_timeout}' defined for check: "
                f"It must be less than the check timeout of {values.data['timeout']} seconds"
            )

        # Return the NTP timeout
        return ntp_timeout



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Manage the master and worker process names
"""

from __future__ import annotations

from typing import Optional

import loguru
from setproctitle import setproctitle


class ProcName:
    """
    Manage the master and worker process names
    """

    def __init__(self, base: str, log_context: loguru.Logger):
        """
        Set up the new ExaCheck sleep object
        """

        # Set the logging context
        self.log = log_context.bind(subsystem="utility")

        # Set the base process name that will be used
        self._base = base

        # Set empty process name string
        self._name = ""

        # Log setup
        self.log.bind(event="debug").info(
            "Created process name manager object with base name '{base}'", base=base
        )

    def update(self, message: str, status: Optional[str] = None) -> None:
        """
        Update the process name with the new message and optionally status
        """
        # Logging
        if status is None:
            self.log.bind(event="debug").trace(
                "Call to set new process name: '{message}'", message=message
            )
        else:
            self.log.bind(event="debug").trace(
                "Call to set new process name with status: '{message} [{status}]'",
                message=message,
                status=status,
            )

        # Combine the base process name with the new process name
        if status is None:
            procname = f"{self._base}: {message}"
        else:
            procname = f"{self._base}: {message} [{status}]"

        # Set the process name
        try:
            setproctitle(procname)
        # pylint: disable=broad-except
        except Exception as exc:
            self.log.bind(event="error").warning(
                "Exception while changing process name: {exception}", exception=exc
            )
        else:
            self.log.bind(event="debug").debug(
                "Set new process name: '{procname}'", procname=procname
            )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

ICMP health check arguments
"""

from typing import Literal, Optional

from pydantic import (
    Field,
    field_validator,
    PositiveInt,
    PositiveFloat,
    NonNegativeInt,
    NonNegativeFloat,
)
from pydantic_core.core_schema import ValidationInfo

from ._remote import Remote


class ICMPArgs(Remote):
    """
    Model for ICMP check arguments
    """

    method: Literal["icmp"] = Field(
        title="ICMP Check Method",
        description="Send ICMP ping packets and ensure the responses are within the configured thresholds",
    )

    count: PositiveInt = Field(
        title="Count",
        description="The number of ICMP packets to send",
        default=3,
    )

    interval: NonNegativeInt | NonNegativeFloat = Field(
        title="Interval",
        description="The interval in seconds between ICMP packets",
        default=0.25,
    )

    icmp_timeout: PositiveInt | PositiveFloat = Field(
        title="ICMP Timeout",
        description="The timeout for each ICMP packet in seconds",
        default=2,
    )

    privileged: bool = Field(
        title="Privileged Mode",
        description="Send ICMP packets in privileged mode (requires root)",
        default=False,
    )

    max_loss: NonNegativeInt = Field(
        title="Maximum Packets Lost",
        description="The maximum number of packets that can be lost before the check is marked as failed",
        default=0,
    )

    max_latency: Optional[PositiveInt | PositiveFloat] = Field(
        title="Maximum Latency",
        description="The highest maximum latency in milliseconds allowed before the check is marked as failed",
        default=None,
    )

    max_jitter: Optional[PositiveInt | PositiveFloat] = Field(
        title="Maximum Jitter",
        description="The maximum jitter in milliseconds allowed before the check is marked as failed",
        default=None,
    )

    @field_validator("count")
    # pylint: disable=no-self-argument
    def validate_count(cls, count: int, values: ValidationInfo) -> int | float:
        """
        Ensure that the count is a valid value

        The assumption is made that the interval will be 1 second; if the interval is higher
        it will be validated as a part of the interval validation
        """
        # Check that count would be below the check timeout
        if count > values.data["timeout"]:
            raise ValueError(
                "Invalid count defined for check: The count must be lower than the check timeout"
            )

        # Return the interval
        return count

    @field_validator("interval")
    # pylint: disable=no-self-argument
    def validate_interval(
        cls, interval: int | float, values: ValidationInfo
    ) -> int | float:
        """
        Ensure that the interval and count is a valid value
        """
        # Get the count
        count: int = values.data["count"] if "count" in values.data else 3

        # Check that count * interval would be below the check timeout
        if (count * interval) > values.data["timeout"]:
            raise ValueError(
                f"Invalid interval '{interval}' or count '{count}' defined for check: "
                f"The total time must be lower than the configured check timeout of {values.data['timeout']} seconds."
            )

        # Return the interval
        return interval

    @field_validator("icmp_timeout")
    # pylint: disable=no-self-argument
    def validate_icmp_timeout(
        cls, icmp_timeout: int | float, values: ValidationInfo
    ) -> int | float:
        """
        Ensure that the ICMP timeout is a valid value

        This is calculate by multiplying the count by the ICMP timeout and checking that the result
        is less than the check timeout
        """
        # Get the variables
        count: int = values.data["count"] if "count" in values.data else 3
        interval: int | float = (
            values.data["interval"] if "interval" in values.data else 1
        )

        # Get the maximum possible time that the check could run if all requests timed out
        maximum_runtime = (count * icmp_timeout) + (interval * (count - 1))

        # Check if the maximum run time is higher than the timeout
        if maximum_runtime > values.data["timeout"]:
            raise ValueError(
                f"Invalid ICMP timeout '{icmp_timeout}' defined for check: "
                f"It must be less than the check timeout of {values.data['timeout']} seconds. "
                "The configured timeouts would result in the check taking too long in a worst case scenario."
            )

        # Return the ICMP timeout
        return icmp_timeout

    @field_validator("max_loss")
    # pylint: disable=no-self-argument
    def validate_max_loss(cls, max_loss: int, values: ValidationInfo) -> int:
        """
        Validate the max loss value is less than the number of packets being sent
        """
        # Get the number of packets to send
        count: int = values.data["count"] if "count" in values.data else 3

        # Verify that the max loss is less than number of packets being sent
        if max_loss >= count:
            raise ValueError(
                f"Invalid max loss '{max_loss}' defined for check: "
                f"It must be less than the number of packets being sent ({count})"
            )

        # Return the max loss
        return max_loss

    @field_validator("max_latency")
    # pylint: disable=no-self-argument
    def validate_max_latency(
        cls, max_latency: int | float, values: ValidationInfo
    ) -> int | float:
        """
        Validate the maximum latency allowed is less than the ICMP timeout
        """
        # Get the ICMP timeout
        timeout: int | float = (
            values.data["icmp_timeout"] if "icmp_timeout" in values.data else 2
        )

        # Verify that the max latency is less than the ICMP timeout
        if max_latency >= timeout:
            raise ValueError(
                f"Invalid max latency '{max_latency}' defined for check: "
                f"It must be less than the ICMP timeout ({timeout})"
            )

        # Return the max latency
        return max_latency

    @field_validator("max_jitter")
    # pylint: disable=no-self-argument
    def validate_max_jitter(
        cls, max_jitter: int | float, values: ValidationInfo
    ) -> int | float:
        """
        If the maximum jitter value is set, require that at least two packets are sent
        """
        # Get the count
        count: int = values.data["count"]

        # Verify more than 1 packet is being sent
        if count == 1:
            raise ValueError(
                f"Invalid max jitter '{max_jitter}' defined for check: "
                f"2 or more packets must be sent to calculate jitter"
            )

        # Return the max jitter
        return max_jitter



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Base check model for any checks that connect to a remote host
"""

from abc import ABC
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Literal

from pydantic import Field, field_validator
from pydantic_core.core_schema import ValidationInfo

from ._base import Base


class Remote(Base, ABC):
    """
    Base model for remote checks which accept a hostname and address family
    """

    host: str = Field(
        title="Hostname/IP Address",
        description="The hostname or IP address the health check will be performed against",
    )

    address_family: Literal["ipv4", "ipv6"] | None = Field(
        title="Address Family",
        description="If checking a hostname, force the check to the specified address family",
        default=None,
    )

    all_valid: bool = Field(
        title="All IP Addresses Valid",
        description=(
            "If checking a hostname, require all IP addresses from the DNS response to pass the health check. "
            "If False, any single IP address that returns a successful check will return the service as healthy."
        ),
        default=False,
    )

    @field_validator("address_family")
    def validate_address_family(  # NOSONAR pylint: disable=no-self-argument
        cls, address_family: Literal["ipv4", "ipv6"], values: ValidationInfo
    ) -> str:
        """
        Validate the address family
        """
        # Try casting the host into an IP address
        try:
            host = ip_address(values.data["host"])
        # pylint: disable=broad-except
        except Exception:
            # The host is not an IP address; nothing to check
            return address_family

        # If the host is an IP address, check that the address family matches
        if isinstance(host, IPv4Address) and address_family == "ipv6":
            raise ValueError("Address family must be IPv4 when querying an IPv4 host")
        if isinstance(host, IPv6Address) and address_family == "ipv4":
            raise ValueError("Address family must be IPv6 when querying an IPv6 host")

        # Return the address family
        return address_family



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Import check argument types
"""

# flake8: noqa

from .dnsargs import DNSArgs
from .fileargs import FileArgs
from .httpargs import HTTPArgs
from .icmpargs import ICMPArgs
from .ntpargs import NTPArgs
from .shellargs import ShellArgs
from .tcpargs import TCPArgs

# Create a Python type of the check args
CheckArgs = DNSArgs | FileArgs | HTTPArgs | ICMPArgs | NTPArgs | ShellArgs | TCPArgs

# Create a Python type of the check args that support remote checks
CheckArgsRemote = DNSArgs | HTTPArgs | ICMPArgs | NTPArgs | TCPArgs



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Manage the sleep between task loop iterations
"""

from __future__ import annotations

from time import monotonic, sleep

import loguru


class Sleeper:
    """
    Manage sleep between task iterations (monitoring or health checks)
    """

    def __init__(self, interval: int | float, log_context: loguru.Logger):
        """
        Set up the new ExaCheck sleep object
        """

        # Set the logging context
        self.log = log_context.bind(subsystem="utility")

        # Set the sleep interval
        self._interval = interval

        # Set empty sleep time type
        self._sleep_time: int | float = 0

        # Log setup
        self.log.bind(event="debug").trace(
            "Created sleep object with interval {interval}", interval=self._interval
        )

        # Start the timer
        self._start = monotonic()
        self.log.bind(event="debug").trace(
            "Setting iteration start time to {start}", start=self._start
        )

    def finish(self) -> None:
        """
        Stop the timer for the current iteration and sleep
        """

        # Set the finish time
        finish = monotonic()
        self.log.bind(event="debug").trace(
            "Setting iteration finish time to {finish}", finish=finish
        )

        # Calculate and set the sleep time
        self._sleep_time = self._calculate(finish=finish)

    @property
    def sleep_time(self) -> int | float:
        """
        Return the expected sleep time
        """
        return self._sleep_time

    def sleep(self) -> None:
        """
        Sleep for the required sleep time
        """
        # Sleep for the required period
        self.log.bind(event="info").info(
            "Sleeping for {sleep_time:.5f} seconds", sleep_time=self._sleep_time
        )
        sleep(self._sleep_time)

        # Wake back up
        self.log.bind(event="info").debug("Waking back up after sleep")

    def _calculate(self, finish: int | float) -> int | float:
        """
        Calculate the sleep time
        """
        # Calculate the elapsed time
        elapsed = finish - self._start
        self.log.bind(event="debug").trace(
            "Calculated iteration took {elapsed:.5f} seconds", elapsed=elapsed
        )

        # Check if the iteration took longer than the interval
        if elapsed > self._interval:
            self.log.bind(event="info").warning(
                "Iteration took longer than the specified sleep interval; sleep time will be set to 1 second"
            )
            return 1

        # Calculate the remaining time left in the interval
        remaining = self._interval - elapsed

        # If the remaining time is less than 1 second, set it to 1 second
        if remaining < 1:
            self.log.bind(event="info").warning(
                "Iteration sleep time is less than 1 second; sleep time will be set to 1 second"
            )
            return 1

        # Return the expected sleep time
        self.log.bind(event="debug").trace(
            "Calculated sleep interval for iteration: {remaining:.5f}",
            remaining=remaining,
        )
        return remaining



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Check Method - File Path
"""


from os import R_OK, access

from ..checkresult import CheckResult
from ..settings.checkargs.fileargs import FileArgs
from ._base import Base


# pylint: disable=too-few-public-methods
class File(Base):
    """
    Run a file path health check
    """

    def check(self) -> CheckResult:  # NOSONAR
        """
        Run the health check
        """
        # Set type for MyPy
        self.args: FileArgs

        # Ensure path can be read
        if not access(self.args.path.parent, R_OK):
            return CheckResult(
                success=False,
                message=(
                    f"File path '{self.args.path}' parent directory '{self.args.path.parent}' is not readable; "
                    "the check will always fail"
                ),
            )

        # Check if the file exists
        try:
            exists = self.args.path.exists()
        except Exception as exc:  # pylint: disable=broad-except
            return CheckResult(
                success=False,
                message=f"Failed to check file '{self.args.path}'",
                error=f"Failed to check file '{self.args.path}': {exc}",
                exception=exc,
            )

        if exists:
            # File exists; check if it should
            if self.args.exists:
                # File should be existing and it does, return the check result
                return CheckResult(
                    success=True,
                    message=f"File {self.args.path} exists",
                )

            # File should not be existing and it does, return
            return CheckResult(
                success=False,
                message=f"File {self.args.path} exists",
                error=f"The file {self.args.path} must not exist",
            )

        # File does not exist, check if it should
        if self.args.exists:
            # File should be existing and it doesn't, return
            return CheckResult(
                success=False,
                message=f"File {self.args.path} does not exist",
                error=f"The file {self.args.path} must exist",
            )

        # File should not be existing and it does, return
        return CheckResult(
            success=True,
            message=f"File {self.args.path} does not exist",
        )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Base check arguments - common to all check types
"""

from abc import ABC
from pprint import pformat

from pydantic import ConfigDict, BaseModel, Field, PositiveInt


class Base(ABC, BaseModel):
    """
    Base model for check arguments
    """

    timeout: PositiveInt = Field(
        title="General Check Timeout",
        description="The total timeout in seconds for the check to execute",
        default=10,
    )
    model_config = ConfigDict(extra="forbid", frozen=True, title="Check Arguments")

    @property
    def pretty(self) -> str:
        """
        Return a pretty printed version of the settings
        """
        return pformat(self.model_dump(), indent=4, width=120)



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Store configuration for notification channels
"""

from typing import Optional, Literal

from pydantic import Field, AnyUrl, field_validator
from pydantic_core.core_schema import ValidationInfo

from ._base import Base


class Notifications(Base):
    name: str = Field(
        title="Notification Name",
        description="The name of the notification service",
        pattern=r"^[^\"']+$",
    )

    description: Optional[str] = Field(
        title="Check Description",
        description="An optional description for the notification service (not parsed)",
        default=None,
    )

    url: AnyUrl = Field(
        title="URL",
        description="The URL to send notifications to (with Apprise)",
    )

    checks: Optional[list[str]] = Field(
        title="Notify Checks",
        description="The list of checks that should have notifications sent to this target",
        default=None,
    )

    events: list[Literal["announce", "info", "error", "withdraw"]] = Field(
        title="Notify Events",
        description="The list of events that should result in notifications being sent to this target",
        default=[
            "announce",
            "error",
            "info",
            "withdraw",
        ],
    )

    general_events: bool = Field(
        title="General Events",
        description="Whether general events not associated with any check should be sent to this target",
        default=False,
    )

    @field_validator("url")
    def validate_url(  # NOSONAR pylint: disable=no-self-argument
        cls, url: AnyUrl, values: ValidationInfo
    ) -> AnyUrl:
        """
        Validate the notification URL
        """
        # Create the apprise object
        import apprise

        apobj = apprise.Apprise()

        # Try adding the notification target
        try:
            apobj.add(url)
        except Exception as exc:
            raise ValueError(f"Invalid notification target URL: {exc}")

        # Create a list to store all schema names for available notifications
        schemas = []

        # Loop over all schemas available from apprise and add to schemas
        details = apobj.details()
        schemas.extend(
            [
                scheme
                for schema in details["schemas"]
                for scheme in schema["details"]["tokens"]["schema"]["values"]
            ]
        )

        # Make sure the URL schema matches
        if url.scheme not in schemas:
            raise ValueError(
                f"Invalid notification target URL scheme (check https://github.com/caronc/apprise for supported schemes): {url.scheme}"
            )

        # Return the URL
        return url



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

DNS health check arguments
"""

from typing import Literal, Optional, Pattern

from pydantic import Field, model_validator, field_validator, PositiveInt, PositiveFloat
from pydantic_core.core_schema import ValidationInfo

from ._remote import Remote


class DNSArgs(Remote):
    """
    Model for DNS check arguments
    """

    method: Literal["dns"] = Field(
        title="DNS Check Method",
        description="Send a query to a DNS server and optionally validate the response",
    )

    query: str = Field(
        title="Query Name",
        description="The name to query",
    )

    query_type: Literal[
        "a", "aaaa", "any", "cname", "mx", "ns", "ptr", "soa", "srv", "txt"
    ] = Field(
        title="Query Type",
        description="The type of DNS query to send",
        default="soa",
    )

    response: Optional[Pattern] = Field(
        title="Response Pattern",
        description="The regular expression to match in the answers received from the DNS server",
        default=None,
    )

    tcp: bool = Field(
        title="Use TCP",
        description="Send the DNS query using TCP rather than UDP",
        default=False,
    )

    port: PositiveInt = Field(
        title="Port",
        description="The port to send the DNS query to",
        default=53,
        le=65535,
    )

    dns_timeout: PositiveInt | PositiveFloat = Field(
        title="DNS Timeout",
        description="The timeout for the DNS request",
        default=5,
    )

    require_resolve: bool = Field(
        title="Require Resolve",
        description="Require the server to response with an answer and not a value such as NXDOMAIN or SRVFAIL",
        default=True,
    )

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_protocol(cls, values: dict) -> dict:
        """
        Convert the protocol to lowercase
        """
        if "protocol" in values and isinstance(values["protocol"], str):
            values["protocol"] = values["protocol"].lower()
        return values

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def set_query_type(cls, values: dict) -> dict:
        """
        Convert the query type to lowercase
        """
        if "query_type" in values and isinstance(values["query_type"], str):
            values["query_type"] = values["query_type"].lower()
        return values

    @field_validator("dns_timeout")
    # pylint: disable=no-self-argument
    def validate_dns_timeout(
        cls, dns_timeout: int | float, values: ValidationInfo
    ) -> int | float:
        """
        Ensure that the DNS timeout is a valid value
        """
        # Check if the DNS timeout is higher than the check timeout
        if dns_timeout > values.data["timeout"]:
            raise ValueError(
                f"Invalid DNS timeout '{dns_timeout}' defined for check: "
                f"It must be less than the check timeout of {values.data['timeout']} seconds"
            )

        # Return the DNS timeout
        return dns_timeout

    @field_validator("require_resolve")
    # pylint: disable=no-self-argument
    def validate_require_resolve(
        cls, require_resolve: bool, values: ValidationInfo
    ) -> bool:
        """
        Validate that require_resolve is set to true if there is a response pattern that needs to be matched
        for the check to be successful
        """
        # Check if a response pattern is defined
        if "response" in values.data and values.data["response"] is not None:
            # Check if require_resolve is false
            if not require_resolve:
                raise ValueError(
                    "Invalid require_resolve value defined for check: "
                    "It must be set to true if a response pattern is defined"
                )

        # Return the require_resolve value
        return require_resolve



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Check Method - DNS
"""

from re import Pattern

import dns.exception as dns_exceptions
from dns import rdatatype, resolver
from dns.rrset import RRset

from ..checkresult import CheckResult
from ..settings.checkargs.dnsargs import DNSArgs
from ._remote import Remote


# pylint: disable=too-few-public-methods,too-many-instance-attributes
class DNS(Remote):
    """
    Run a DNS health check
    """

    def check_one(self, addr: str) -> CheckResult:  # NOSONAR
        """
        Run the health check
        """
        # Set type for MyPy
        self.args: DNSArgs

        # Send request to the DNS server
        try:
            response = self._get_dns(addr=addr)
        except dns_exceptions.Timeout:
            # The query timed out; return failed check result
            return CheckResult(
                success=False,
                message=f"DNS request to {self.args.host}:{self.args.port} timed out",
                output=f"Connection was initiated to IP {addr}:{self.args.port}",
                error="DNS query timed out",
            )
        except dns_exceptions.DNSException as exc:
            # Handle the exception that was received
            return self._handle_dns_exception(addr=addr, exc=exc)

        # Process the response into a CheckResult
        return self._process_dns_response(addr=addr, response=response)

    def _process_dns_response(
        self, addr: str, response: resolver.Answer
    ) -> CheckResult:
        """
        Process a DNS response to ensure it is valid
        """
        self.log.bind(event="debug").debug(
            "Processing DNS response to validate against check parameters"
        )

        # Ensure there is an rrset if required
        if not response.rrset and self.args.require_resolve:
            return CheckResult(
                success=False,
                message=f"DNS query to {self.args.host}:{self.args.port} returned no answers",
                output=f"The query response from IP address {addr} returned no answer for the requested name",
            )

        # If there is no pattern to match, return a successful check result
        if not self.args.response:
            return CheckResult(
                success=True,
                message=f"DNS query to {self.args.host}:{self.args.port} returned a response",
                output=f"The query response from IP address {addr} returned a response for the requested name",
            )

        # Search for the response pattern in response and return
        return self._search_response_pattern(addr=addr, response=response)

    def _search_response_pattern(
        self, addr: str, response: resolver.Answer
    ) -> CheckResult:
        """
        Search for a pattern in the DNS response
        """
        # Set type for MyPy
        assert isinstance(response.rrset, RRset)
        assert isinstance(self.args.response, Pattern)

        # Loop over each answer in the response and look for the response pattern
        self.log.bind(event="debug").debug(
            "Searching DNS response for a match to response pattern: {pattern}",
            pattern=self.args.response.pattern,
        )
        for answer in [answer.to_text() for answer in response.rrset]:
            # Test if the answer matches
            self.log.bind(event="debug").trace(
                "Testing answer: {answer}",
                answer=answer,
            )
            if self.args.response.match(answer):
                # Match found
                return CheckResult(
                    success=True,
                    message=f"DNS query to {self.args.host}:{self.args.port} returned a validated response",
                    output=(
                        f"The query response from IP address {addr} returned a response with content matching "
                        f"response pattern: {answer}"
                    ),
                )

            # No match found, continue to next answer if any
            self.log.bind(event="debug").debug(
                "Answer did not match response pattern: {answer}",
                answer=answer,
            )

        # No match found, check failed
        return CheckResult(
            success=False,
            message=f"DNS query to {self.args.host}:{self.args.port} returned a response without matches",
            output=(
                f"The query response from IP address {addr} returned a response but no matches were found for the "
                f"pattern {self.args.response}"
            ),
        )

    def _handle_dns_exception(  # NOSONAR
        self, addr: str, exc: dns_exceptions.DNSException
    ) -> CheckResult:
        """
        Handle a DNS exception
        """
        # Log the exception at a low level; it could mean nothing at all without further checking
        self.log.bind(event="debug").debug(
            "DNS exception while sending {query_type} query to {host}:{port} with IP {addr} for name {query}",
            query_type=self.args.query_type,
            host=self.args.host,
            port=self.args.port,
            addr=addr,
            query=self.args.query,
        )
        self.log.bind(event="debug").trace(
            "DNS exception:\n{exc}",
            exc=exc,
        )

        # Check the type of exception
        match type(exc):
            # NXDOMAIN and NoAnswer indicate that a response was received; check if a response is required
            # and if it needs to be validated
            case resolver.NXDOMAIN | resolver.NoAnswer:
                # Test if the response should be validated
                if self.args.require_resolve or self.args.response:
                    return CheckResult(
                        success=False,
                        message=f"DNS query to {self.args.host}:{self.args.port} returned no answer",
                        output=f"The query response from IP address {addr} returned no answer for the requested name",
                        error=f"{exc}",
                        exception=exc,
                    )

            # YXDOMAIN and NoNameservers indicate that a response was received but it is invalid
            case resolver.YXDOMAIN | resolver.NoNameservers:
                # Test if the response should be validated
                if self.args.require_resolve or self.args.response:
                    return CheckResult(
                        success=False,
                        message=f"DNS query to {self.args.host}:{self.args.port} returned an invalid response",
                        output=f"The query response from IP address {addr} returned an invalid response",
                        error=exc.msg,
                        exception=exc,
                    )

            # LifetimeTimeout indicates the query timed out
            case resolver.LifetimeTimeout:
                return CheckResult(
                    success=False,
                    message=f"DNS query to {self.args.host}:{self.args.port} timed out",
                    output=f"No query response was received from IP address {addr}",
                    error=f"{exc}",
                    exception=exc,
                )

            # Catch other exceptions
            case _:
                return CheckResult(
                    success=False,
                    message=f"DNS query to {self.args.host}:{self.args.port} returned an exception",
                    output=f"The DNS query to IP address {addr} returned an unhandled exception",
                    error=f"{exc}",
                    exception=exc,
                )

        # If reaching this point the exception was not fatal; return the check result
        return CheckResult(
            success=True,
            message=f"DNS query to {self.args.host}:{self.args.port} succeeded",
            output=f"The query response from IP address {addr} an answer which did not require validation",
            error=f"{exc}",
            exception=exc,
        )

    def _create_client(self, addr: str) -> resolver.Resolver:
        """
        Create a DNS resolver object
        """
        # Create the resolver object
        client = resolver.Resolver(configure=False)

        # Set the timeout for the query
        client.timeout = self.args.timeout
        client.lifetime = self.args.timeout

        # Set the name server that will be queried
        client.nameservers = [addr]

        # Set the port to query
        client.port = self.args.port

        return client

    def _get_dns(self, addr: str) -> resolver.Answer:
        """
        Send a DNS query to the specified host
        """
        # Create the resolver object
        client = self._create_client(addr=addr)

        # Log query about to happen
        self.log.bind(event="debug").debug(
            "Sending {query_type} query to {addr}:{port} for name {query}",
            query_type=self.args.query_type,
            addr=addr,
            port=self.args.port,
            query=self.args.query,
        )

        # Send the DNS query
        result = client.resolve(
            qname=self.args.query,
            rdtype=rdatatype.from_text(self.args.query_type),
            tcp=self.args.tcp,
            raise_on_no_answer=True,
        )
        self.log.bind(event="debug").debug(
            "DNS response received from {addr}:{port}",
            addr=addr,
            port=self.args.port,
        )

        # Dump response if available
        if result.rrset:
            self.log.bind(event="datadump").trace(
                "Raw DNS response:\n{response}",
                response=f"{result.rrset.to_text()}",
            )
        else:
            self.log.bind(event="debug").debug(
                "Empty DNS response received from {addr}:{port}"
            )

        # Return the answer
        return result



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Load and manage the configuration for ExaCheck
"""

from __future__ import annotations

from pathlib import Path
from pprint import pformat
from typing import Optional
import sys

import ujson
import loguru
from pydantic import ValidationError
from yaml import safe_load as yaml_safe_load

from .settings.settings import Settings


class Configuration:
    """
    ExaCheck configuration handler
    """

    def __init__(
        self,
        log_context: loguru.Logger,
        file: Optional[Path] = None,
        configuration: Optional[dict] = None,
    ):
        # Set the logging context
        self.log = log_context.bind(subsystem="configuration")

        # Ensure that either a file or configuration dict is provided
        if not file and not configuration:
            self.log.bind(event="error").critical(
                "No configuration file or configuration dict provided"
            )
            raise SystemExit(1)

        # Check if a file name was provided
        if file:
            # Ensure the file name is yaml or json
            if file.suffix not in (".yaml", ".yml", ".json"):
                self.log.bind(event="error").critical(
                    "Configuration file '{file}' must have a '.json' or '.yaml' extension",
                    file=file,
                )
                raise SystemExit(1)

            # Read the configuration file into a dict
            configuration = self._load_file(file=file)

            # Set the modification time for the file
            self.mtime = file.stat().st_mtime

        # Parse the configuration into a Settings object
        try:
            self.settings = Settings(**configuration, file=file)
        except ValidationError as exc:
            # Invalid configuration
            self.log.bind(event="error").critical(
                "The configuration could not be parsed due to validation errors."
            )
            for error in exc.errors():
                self.log.bind(event="error").error(
                    "{location}: {error}",
                    location=error["loc"],
                    error=error["msg"],
                )
            self.log.bind(event="datadump").error(
                "Pydantic reported the following errors:\n{errors}",
                errors=pformat(exc.errors()),
            )
            # Exit
            sys.exit(2)

        except Exception as exc:
            # An unexpected exception occurred; raise SystemExit exception to exit
            raise SystemExit(99) from exc

        # Return loaded settings
        self.log.bind(event="info").info("Configuration file loaded successfully")
        self.log.bind(event="datadump").trace(
            "Loaded settings from configuration:\n{settings}",
            settings=self.settings.pretty,
        )

    def _load_file(self, file: Path) -> dict:
        """
        Read configuration from the supplied JSON or YAML file into a dict
        """
        # Logging
        self.log.bind(event="info").debug(
            "Loading configuration data from file '{file}'",
            file=file,
        )

        # Check the file type
        match file.suffix:
            case file.suffix if file.suffix in (".yaml", ".yml"):
                configuration = yaml_safe_load(file.read_text(encoding="utf-8"))
            case ".json":
                configuration = ujson.loads(file.read_text(encoding="utf-8"))
            case _:
                self.log.bind(event="error").critical(
                    "Configuration file '{file}' does not have a valid extension",
                    file=file,
                )
                raise SystemExit(1)

        # Dump the configuration dict
        self.log.bind(event="datadump").trace(
            "Read {file_type} data:\n{file_content}",
            file_type=file.suffix[1:].upper(),
            file_content=pformat(configuration, indent=4, width=120),
        )

        # Return the configuration dict
        return configuration

    def is_modified(self) -> bool:
        """Check if the configuration file has been modified"""
        # Configuration can only be modified if it is from a configuration file
        if not self.settings.file:
            return False

        # Logging
        self.log.bind(event="debug").trace(
            "Testing if configuration file has been modified",
        )

        # Get the current modification time of the configuration file
        mtime = self.settings.file.stat().st_mtime

        # Compare the modification times
        if mtime != self.mtime:
            # Logging
            self.log.bind(event="info").info(
                "Configuration file has been modified",
            )

            # Return True
            return True

        # Configuration not modified
        self.log.bind(event="debug").trace(
            "Configuration file has not been modified",
        )
        return False

    def reload(self) -> bool:
        """Reload the configuration file"""
        # Configuration file appears to have been modified, attempt to parse it into a dict
        try:
            configuration = self._load_file(file=self.settings.file)
        except Exception as exc:
            # Log failure
            self.log.bind(event="error").error(
                "Configuration file has been modified but could not be parsed: {exc}",
                exc=exc,
            )
            # Update the modification time
            self.mtime = self.settings.file.stat().st_mtime
            # Return False to indicate that the configuration was not reloaded
            return False

        # Configuration file was modified and parsed successfully, attempt to load it
        try:
            settings = Settings(**configuration, file=self.settings.file)
        except ValidationError as exc:
            # Log failure
            self.log.bind(event="error").error(
                "Configuration file is invalid: {exc}",
                exc=exc,
            )
            # Update the modification time
            self.mtime = self.settings.file.stat().st_mtime
            # Return False to indicate that the configuration was not reloaded
            return False
        except Exception as exc:
            # Log failure
            self.log.bind(event="error").error(
                "Exception loading configuration file: {exc}",
                exc=exc,
            )
            # Update the modification time
            self.mtime = self.settings.file.stat().st_mtime
            # Return False to indicate that the configuration was not reloaded
            return False

        # Update the modification time
        self.mtime = self.settings.file.stat().st_mtime

        # Configuration file was loaded successfully, update the settings
        self.settings = settings

        # Log success
        self.log.bind(event="info").info(
            "The configuration has been reloaded successfully",
        )

        # Return True to indicate that the configuration was reloaded
        return True



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Check Method - ICMP
"""

from __future__ import annotations

from icmplib import SocketPermissionError, TimeoutExceeded, ping, Host

from ..checkresult import CheckResult
from ..settings.checkargs.icmpargs import ICMPArgs
from ._remote import Remote


# pylint: disable=too-few-public-methods
class ICMP(Remote):
    """
    Run a ICMP health check
    """

    def check_one(self, addr: str) -> CheckResult:
        """
        Run the ICMP health check
        """
        # Set type for MyPy
        self.args: ICMPArgs

        # Send the ping/ICMP request and get response
        try:
            response = self._ping(addr=addr)
        except SocketPermissionError as exc:
            # If a SocketPermissionError is raised, then the user does not have sufficient permissions; the check
            # will fail for all addresses so return a failure immediately
            self.log.bind(event="error").error(
                (
                    "Ping to host {host} failed due to insufficient permissions. "
                    "The service is marked as failed due to this error. The ping request was sent to IP {addr}."
                ),
                host=self.args.host,
                addr=addr,
            )
            return CheckResult(
                success=False,
                message="Failed to create socket; insufficient permissions",
                output=f"Ping request destined to '{self.args.host}' failed due to insufficient permissions",
                error=f"Exception raised: {exc}",
                exception=exc,
            )
        except TimeoutExceeded as exc:
            # If there was a timeout exceeded exception the check can't have further processing performed
            self.log.bind(event="error").warning(
                (
                    "Ping to {host} using IP {addr} raised timeout; returning failure for check as all "
                    "results must be valid"
                ),
                host=self.args.host,
                addr=addr,
            )
            return CheckResult(
                success=False,
                message=f"Timeout exceeded pinging host {self.args.host}",
                output=f"Ping was sent to IP {addr}",
                error=f"{exc}",
            )

        # Process the response into a check result and return
        return self._process_icmp_response(response=response)

    def _ping(self, addr: str) -> Host:
        """
        Ping an IP address and retrieve the result
        """
        # Attempt to ping the provided IP address
        self.log.bind(event="debug").debug(
            "Sending {count} ping requests to {addr} using {privileged} mode",
            count=self.args.count,
            addr=addr,
            privileged="privileged" if self.args.privileged else "unprivileged",
        )
        result = ping(
            address=addr,
            count=self.args.count,
            interval=self.args.interval,  # type: ignore
            timeout=self.args.timeout,
            privileged=self.args.privileged,
        )

        # Log the result
        self.log.bind(event="debug").debug(
            "Ping response indicates that {addr} is {reachable}",
            addr=addr,
            reachable="reachable" if result.is_alive else "unreachable",
        )
        self.log.bind(event="datadump").trace(
            "Raw ping result:\n{response}",
            response=f"{result}",
        )

        # Return the result
        return result

    def _process_icmp_response(self, response: Host) -> CheckResult:
        """
        Process the results of an ICMP test to ensure the values are within the expected parameters
        """
        self.log.bind(event="debug").debug(
            "Processing ICMP response to validate against check parameters",
        )

        # Ensure that at least one response is received
        if response.packets_received == 0:
            # Create log of failure
            self.log.bind(event="info").warning(
                "Ping request to {host} with IP {addr} failed; no packets received",
                host=self.args.host,
                addr=response.address,
            )

            # Build result
            result = CheckResult(
                success=False,
                message=f"No ICMP response from {self.args.host} with IP {response.address}",
            )

            # Return response early
            self.log.bind(event="debug").debug(
                "Returning CheckResult with failure due to missing ICMP response"
            )
            self.log.bind(event="datadump").trace(
                "Raw check result:\n{result}",
                result=result.pretty,
            )
            return result

        # Build output message
        output = (
            f"Packets:{self.args.count} Received:{response.packets_received} "
            f"Loss:{response.packet_loss} Interval:{self.args.interval} Min:{response.min_rtt}ms "
            f"Max:{response.max_rtt}ms Avg:{response.avg_rtt}ms"
        )

        # Create list of errors that may be returned
        errors: list[str] = []

        # Set default success status to True
        success = True

        # Check if the number of packets lost is within the allowed range
        if (
            lost := response.packets_sent - response.packets_received
            > self.args.max_loss
        ):
            success = False
            message = (
                f"Packets lost ({lost}) exceeds maximum limit of {self.args.max_loss}"
            )
            self.log.bind(event="debug").error(message)
            errors.append(message)

        # Check if the latency (if set) is within the allowed range
        if self.args.max_latency and response.max_rtt > self.args.max_latency:
            success = False
            message = f"Maximum latency ({response.max_rtt}ms) exceeds maximum limit of {self.args.max_latency}ms"
            self.log.bind(event="debug").error(message)
            errors.append(message)

        # Check if the jitter (if set) is within the allowed range
        if (
            self.args.max_jitter
            and response.packets_received > 1
            and response.jitter > self.args.max_jitter
        ):
            success = False
            message = f"Maximum jitter ({response.max_rtt}ms) exceeds limit of {self.args.max_jitter}ms"
            self.log.bind(event="debug").error(message)
            errors.append(message)

        # If there were errors, build the error string
        error: str | None
        if errors:
            error = ", ".join(errors)
        else:
            error = None

        # Build message
        if success:
            message = f"ICMP test to {response.address} successful"
        else:
            message = f"ICMP test to {response.address} failed"

        # Create the check result object and return
        return CheckResult(
            success=success,
            message=message,
            output=output,
            error=error,
        )



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Check Method - TCP Connection
"""

from __future__ import annotations

import socket
from ipaddress import ip_address, IPv4Address

from ..checkresult import CheckResult
from ..settings.checkargs.tcpargs import TCPArgs
from ._remote import Remote


# pylint: disable=too-few-public-methods
class TCP(Remote):
    """
    Run a health check to a TCP port to ensure it is listening
    """

    def check_one(self, addr: str) -> CheckResult:
        """
        Run a TCP health check to the supplied IP address
        """
        # Set type for MyPy
        self.args: TCPArgs

        # Create IP address object
        ipaddr = ip_address(addr)

        # Set the address family type
        family = socket.AF_INET if isinstance(ipaddr, IPv4Address) else socket.AF_INET6

        # Run the health check and check if successful
        if self._test_socket(addr=addr, family=family):
            # Return success
            return CheckResult(
                success=True,
                message=f"TCP connection to {self.args.host}:{self.args.port} succeeded",
                output=f"Connection was initiated to IP {addr}:{self.args.port}",
            )

        # Failure opening the connection
        return CheckResult(
            success=False,
            message=f"TCP connection to {self.args.host}:{self.args.port} failed",
            output=f"Connection was initiated to IP {addr}:{self.args.port}",
        )

    def _test_socket(self, addr: str, family) -> bool:
        """
        Attempt making a TCP connection to the IP address/port and return true/false
        """

        # Set up the socket
        self.log.bind(event="debug").debug(
            "Opening socket to {addr}:{port}",
            addr=addr,
            port=self.args.port,
        )
        sock = socket.socket(family=family, type=socket.SOCK_STREAM)
        sock.settimeout(self.args.tcp_timeout)

        # Open connection
        self.log.bind(event="debug").debug(
            "Testing connection to {addr}:{port}",
            addr=addr,
            port=self.args.port,
        )
        try:
            result = sock.connect_ex((addr, self.args.port))
        except Exception as exc:  # pylint: disable=broad-except
            # Log failure
            self.log.bind(event="debug").debug(
                "Error connecting to {addr}:{port}: {error}",
                addr=addr,
                port=self.args.port,
                error=f"{exc}",
                exception=exc,
            )

            # Shut down/close socket
            try:
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()
            except Exception:  # pylint: disable=broad-except
                pass

            # Return failed result
            return False

        # Shut down/close socket
        try:
            self.log.bind(event="debug").debug(
                "Shutting down socket to {addr}:{port}",
                addr=addr,
                port=self.args.port,
            )
            sock.shutdown(socket.SHUT_RDWR)
            sock.close()
        except Exception:  # pylint: disable=broad-except
            pass

        # Return the result as a bool
        self.log.bind(event="debug").debug(
            "Connection to {addr}:{port} returned code: {result}",
            addr=addr,
            port=self.args.port,
            result=result,
        )
        return result == 0



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Check Method - NTP
"""

from __future__ import annotations

from datetime import datetime

from ntplib import NTPClient, NTPStats, leap_to_text, ref_id_to_text
from tabulate import tabulate

from ..checkresult import CheckResult
from ..settings.checkargs.ntpargs import NTPArgs
from ._remote import Remote


# pylint: disable=too-few-public-methods
class NTP(Remote):
    """
    Run a NTP health check
    """

    def check_one(self, addr: str) -> CheckResult:
        """
        Run the health check
        """
        # Set type for MyPy
        self.args: NTPArgs

        # Send the NTP request and get response
        try:
            response = self._get_ntp(addr=addr)
        except Exception as exc:  # pylint: disable=broad-except
            self.log.bind(event="debug").error(
                "NTP connection exception: {exc}",
                exc=f"{exc}",
            )
            # Return a failure as the check raised exception
            return CheckResult(
                success=False,
                message=f"NTP connection to {self.args.host}:{self.args.port} failed",
                output=f"Failed connection to IP {addr}:{self.args.port}",
                error=f"Exception raised: {exc}",
                exception=exc,
            )

        # Process the response into a check result and return
        return self._process_ntp_response(addr=addr, response=response)

    def _get_ntp(self, addr: str) -> NTPStats:
        """
        Get an NTP response from the specified address
        """
        # Create the NTP client object
        client = NTPClient()

        # Send the NTP request and get response
        self.log.bind(event="debug").debug(
            "Sending NTP request to {addr}:{port} using NTP v{version}",
            addr=addr,
            port=self.args.port,
            version=self.args.version,
        )
        response = client.request(
            host=addr,
            version=self.args.version,
            port=f"{self.args.port}",
            timeout=self.args.timeout,
        )
        self.log.bind(event="debug").debug(
            "NTP response received with offset: {offset}",
            offset=response.offset,
        )
        self.log.opt(lazy=True).bind(event="datadump").trace(
            "NTP data received:\n{result_table}",
            result_table=lambda: self._result_table(response),
        )

        # Return the NTP response
        return response

    def _process_ntp_response(self, addr: str, response: NTPStats) -> CheckResult:
        """
        Process the NTP response into a CheckResult object
        """
        self.log.bind(event="debug").debug(
            "Processing NTP response to validate against check parameters"
        )

        # Get the various statistics from response
        offset = response.offset
        reference = ref_id_to_text(response.ref_id)
        leap = leap_to_text(response.leap)
        version = response.version
        stratum = response.stratum

        # Create list of errors that may be returned
        errors: list[str] = []

        # Build output message
        output = (
            f"IP:{addr} Port:{self.args.port} Version:{version} Stratum:{stratum} Offset:{offset} "
            f"Reference:{reference} Leap Status:{leap}"
        )

        # Set success status to default to True; if failure is detected, this will be set to False
        success = True

        # Check if the offset is within the maximum offset value (+/-)
        if isinstance(self.args.max_offset, (int, float)):
            if abs(offset) > self.args.max_offset:
                # Offset is +/- over maximum offset
                success = False
                message = f"Offset {offset} exceeds maximum offset of {self.args.max_offset} compared to local time"
                self.log.bind(event="debug").error(message)
                errors.append(message)

        # Check if the stratum is within the maximum stratum value
        if isinstance(self.args.max_stratum, int):
            if stratum > self.args.max_stratum:
                # NTP server stratum is too high
                success = False
                message = f"Stratum {stratum} exceeds maximum stratum of {self.args.max_stratum}"
                self.log.bind(event="debug").error(message)
                errors.append(message)

        # Ensure the response version matches the requested version
        if self.args.version != version:
            success = False
            message = f"Response version {version} does not match requested version {self.args.version}"
            self.log.bind(event="debug").error(message)
            errors.append(message)

        # If there were errors, build the error string
        error: str | None
        if errors:
            error = ", ".join(errors)
        else:
            error = None

        # Build message
        if success:
            message = f"NTP check to {addr}:{self.args.port} successful"
        else:
            message = f"NTP check to {addr}:{self.args.port} failed"

        # Generate and return the check result
        return CheckResult(
            success=success,
            message=message,
            output=output,
            error=error,
        )

    @staticmethod
    def _result_table(result: NTPStats) -> str:
        """
        Generate a table of result information from the NTP response
        """
        # Set headers for the table
        headers = ["Name", "Value"]

        # Create data list
        data: list[list[str]] = []

        # Add the data to the data list
        data.append(["Offset", f"{result.offset}"])
        data.append(["Reference Clock", ref_id_to_text(result.ref_id)])
        data.append(["Stratum", f"{result.stratum}"])
        data.append(["Leap Status", leap_to_text(result.leap)])
        data.append(["Version", f"{result.version}"])
        data.append(["Precision", f"{result.precision}"])
        data.append(["TX Time", f"{result.tx_time}"])
        data.append(["RX Time", f"{result.recv_time}"])
        data.append(["TX Time Human", f"{datetime.fromtimestamp(result.tx_time)}"])
        data.append(["RX Time Human", f"{datetime.fromtimestamp(result.recv_time)}"])
        data.append(["Root Delay", f"{result.root_delay}"])

        # Create the table
        table = tabulate(data, headers=headers)

        # Return the table
        return table



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Check timeout exception
"""


class CheckTimeout(Exception):
    """
    Exception raised when a check times out
    """

    def __init__(self, message: str = "The check timed out"):
        """
        Set up the new CheckTimeout object
        """
        self.message = message

    def __str__(self):
        return self.message



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Store configuration for the ExaCheck application.
"""

from typing import Optional, Annotated, Union

from pydantic import (
    Field,
    model_validator,
    field_validator,
    FilePath,
)
from pydantic_core.core_schema import ValidationInfo

from ._base import Base
from .check import Check
from .loggers import LogFile, Syslog
from .sentry import Sentry
from .notifications import Notifications
from .exacheck import ExaCheck


# Set the available log types to allow the log type to be discriminated
LogMethods = Annotated[Union[LogFile, Syslog], Field(discriminator="method")]


class Settings(Base):
    """
    Define and store the parsed configuration for ExaCheck
    """

    # File path set to optional as it is defined by the CLI
    file: Optional[FilePath] = Field(
        title="Configuration File",
        description="The path to the configuration file that was loaded",
        default=None,
    )

    exacheck: ExaCheck = Field(
        title="ExaCheck Configuration",
        description="ExaCheck internal options",
        default=ExaCheck(),
    )

    checks: list[Check] = Field(
        title="Checks",
        description="The list of health checks to perform",
    )

    logging: Optional[list[LogMethods]] = Field(
        title="Logging Configuration",
        description="The logging configuration for ExaCheck",
        default=None,
    )

    notifications: Optional[list[Notifications]] = Field(
        title="Notifications",
        description="The list of services to send notifications to for check events",
        default=None,
    )

    sentry: Optional[Sentry] = Field(
        title="Sentry Configuration",
        description="The Sentry SDK configuration for error reporting/profiling",
        default=None,
    )

    @model_validator(mode="before")
    # pylint: disable=no-self-argument
    def validate_logfiles_unique(cls, values: dict) -> dict:
        """
        If any file based loggers are defined, ensure the file names are unique
        """
        # Skip if no logging defined
        if "logging" not in values or values["logging"] is None:
            return values

        # Get a list of all log file paths from file loggers
        paths = [
            f"{logger.destination}"
            for logger in values["logging"]
            if isinstance(logger, LogFile)
        ]

        # Create set of unique paths for searching duplicates
        unique_paths = set()

        # Search for duplicates
        duplicates = [path for path in paths if path in unique_paths or unique_paths.add(path)]  # type: ignore

        # Raise error if duplicates found
        if duplicates:
            raise ValueError(
                f"Log files must be unique. Duplicate log file paths found: {', '.join(duplicates)}"
            )

        # Return the validated values
        return values

    @field_validator("logging")
    # pylint: disable=no-self-argument
    def validate_log_filter_names(
        cls, logging: list[LogFile | Syslog], values: ValidationInfo
    ) -> list[LogFile | Syslog]:
        """
        If any check name log filters are defined, ensure the check name actually exists
        """
        # Create a list of all check names
        check_names: list[str] = [check.name for check in values.data["checks"]]

        # Loop through each logger
        for logger in logging:
            # Skip if no check filters defined
            if not logger.checks:
                continue

            # Check if undefined check names have been set
            if undefined := [
                check_name
                for check_name in logger.checks
                if check_name not in check_names
            ]:
                if logger.method == "file":
                    raise ValueError(
                        f"Log to file {logger.destination} has undefined check name(s) configured: "
                        f"{', '.join(undefined)}"
                    )
                raise ValueError(
                    f"Log to syslog server {logger.destination} has undefined check name(s) configured: "
                    f"{', '.join(undefined)}"
                )

        # Return the validated data
        return logging

    def get_check_name(self, name: str) -> Check | None:
        """Retrieve a check object by name"""
        # Loop through each check searching for the matching name
        for check in self.checks:
            if check.name == name:
                return check



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Worker process creation error
"""


class WorkerProcessError(Exception):
    """
    Exception raised when a worker process could not be created
    """

    def __init__(self, message: str = "Exception creating worker process"):
        """
        Set up the new WorkerProcessError object
        """
        self.message = message

    def __str__(self):
        return self.message



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Import check methods
"""

from .dns import DNS
from .file import File
from .http import HTTP
from .icmp import ICMP
from .ntp import NTP
from .shell import Shell
from .tcp import TCP

# Create a Python type of the check methods
CheckMethods = DNS | File | HTTP | ICMP | NTP | Shell | TCP

# Create a list of check methods with their class
METHODS = {
    "dns": ("DNS", DNS),
    "file": ("File", File),
    "http": ("HTTP", HTTP),
    "icmp": ("ICMP", ICMP),
    "ntp": ("NTP", NTP),
    "shell": ("Shell", Shell),
    "tcp": ("TCP", TCP),
}



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Base logging configuration class
"""

from __future__ import annotations

from abc import ABC
from typing import Literal, Optional

from pydantic import ConfigDict, BaseModel, Field, PrivateAttr

from .._base import Base as BaseSettings


class Base(BaseSettings, ABC):
    """
    Log settings common to both file logging and syslog
    """

    # Create class to store the live logging configuration in a private attribute (_conf)
    class LiveConfiguration(BaseModel):
        """
        Store the live logger configuration
        """

        id: int | None = Field(
            title="Logger ID",
            description="The logger ID for this logging target",
            default=None,
        )
        model_config = ConfigDict(extra="forbid")

    level: Literal["error", "warning", "info", "success", "debug", "trace"] = Field(
        title="Log Level",
        description="The minimum log level to write logs for",
        default="info",
    )

    formatter: Optional[str] = Field(
        title="Custom Format String",
        description="The custom log format string to use",
        default=None,
    )

    events: list[
        Literal["announce", "datadump", "debug", "error", "info", "withdraw"]
    ] = Field(
        title="Log Events",
        description="The list of events that should be logged to this logging target",
        default=[
            "announce",
            "error",
            "info",
            "withdraw",
        ],
    )

    subsystems: list[
        Literal[
            "announcer",
            "configuration",
            "executor",
            "healthcheck",
            "logging",
            "master",
            "notification",
            "utility",
            "worker",
        ]
    ] = Field(
        title="Log Subsystem",
        description="The list of subsystems that should be logged to this logging target",
        default=[
            "announcer",
            "configuration",
            "executor",
            "healthcheck",
            "logging",
            "master",
            "notification",
            "utility",
            "worker",
        ],
    )

    checks: Optional[list[str]] = Field(
        title="Log Checks",
        description="The list of checks that should be logged to this logging target",
        default=None,
    )

    _conf: LiveConfiguration = PrivateAttr(default=LiveConfiguration())

    def set_logger_id(self, logger_id: int) -> None:
        """Set the logger ID for this logging target

        Args:
            logger_id (int): The logger ID
        """
        self._conf.id = logger_id

    def get_logger_id(self) -> int | None:
        """
        Get the logger ID for this logging target
        """
        return self._conf.id



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Base Check Method for remote checks
"""

from __future__ import annotations

import socket
from abc import ABC, abstractmethod
from enum import Enum
from pprint import pformat
from typing import Literal
from ipaddress import ip_address, IPv4Address

from ..settings.checkargs import CheckArgsRemote
from ..exceptions.dnsresolutionerror import DNSResolutionError
from ..exceptions.addressfamilyerror import AddressFamilyError
from ..checkresult import CheckResult
from ._base import Base


class Remote(Base, ABC):
    """
    The remote base health check class
    """

    class AddressFamilies(Enum):
        """
        Create enum for lookups of address families from the string IPv4/IPv6 to a socket object
        """

        IPV4 = socket.AF_INET
        IPV6 = socket.AF_INET6
        UNSPECIFIED = socket.AF_UNSPEC

        @classmethod
        def _missing_(cls, value):
            """
            Capitalize the value provided if missing from enum and try to match it to a member
            """
            # First check there was even a value provided; if not, return the default value of unspecified
            if not value:
                return cls.UNSPECIFIED

            # Set type for MyPy to string
            assert isinstance(value, str)

            # Search through enum
            for member in cls:
                if member.name == value.upper():
                    return member

            # Return the default value of unspecified if not defined
            return cls.UNSPECIFIED

    def __init__(self, *args, **kwargs):
        """
        Add address family definition to init; used to resolve the host to an IP address
        """
        # Call the parent init
        super().__init__(*args, **kwargs)

        # Define the address family to use for DNS resolution
        self.address_family = self.AddressFamilies(self.args.address_family).value  # type: ignore

    def check(self) -> CheckResult:
        """
        Loop over each IP address and perform the health check
        """
        # Inform MyPy that the check args are for a remote check
        self.args: CheckArgsRemote
        assert isinstance(self.args.host, str)

        # Resolve the host into a list of IP addresses
        ips = self.resolve(
            host=self.args.host,
        )

        # Loop over each IP address
        for addr in ips:
            # Run the individual health check and return if a check result was retrieved
            if (result := self._check(addr=addr)) is not None:
                return result

        # If all health checks were required to be valid, return a success result
        if self.args.all_valid:
            self.log.bind(event="info").success(
                "Health check for {host} using all resolved IP address was successful; tested IP addresses: {ips}",
                host=self.args.host,
                ips=", ".join(ips),
            )
            # Return a generated check result
            return CheckResult(
                success=True,
                message="Health check for all resolved IP addresses successful",
                output=f"Tested IP addresses: {', '.join(ips)}",
            )

        # If reaching this point generate and return failed check result
        self.log.bind(event="error").warning(
            "Health check for {host} failed for one or more IP addresses; tested IP addresses: {ips}",
            host=self.args.host,
            ips=", ".join(ips),
        )
        return CheckResult(
            success=False,
            message="Health check for one or more IP addresses failed",
            output=f"Tested IP addresses: {', '.join(ips)}",
        )

    def _check(self, addr: str) -> CheckResult | None:
        """
        Perform the health check to a single address
        """
        # Log the check start
        self.log.bind(event="debug").debug(
            "Health check started for host {host} using IP address {addr}",
            host=self.args.host,
            addr=addr,
        )

        # Run the health check
        try:
            result = self.check_one(addr=addr)
        except Exception as exc:  # pylint: disable=broad-except
            self.log.bind(event="debug").error(
                "Health check raised unhandled exception: {exc}",
                exc=f"{exc}",
            )
            result = CheckResult(
                success=False,
                message="Check raised unhandled exception",
                error=f"Unhandled exception raised for IP address {addr}: {exc}",
                exception=exc,
            )

        # Log the result
        self._log_result(result=result)

        # If all health checks must be successful and a failure was returned, return the failed check result
        if self.args.all_valid and result.success is False:
            self.log.bind(event="error").warning(
                "Health check for {host} using IP address {addr} failed; skipping subsequent IPs if any",
                host=self.args.host,
                addr=addr,
            )
            return result

        # If only a single health check needs to be successful, return the successful result
        if not self.args.all_valid and result.success is True:
            self.log.bind(event="info").success(
                "Health check for {host} using IP address {addr} successful; skipping subsequent IPs if any",
                host=self.args.host,
                addr=addr,
            )
            return result

        # If the check was successful, log but continue on to next health check
        if result.success:
            self.log.bind(event="info").success(
                (
                    "Health check for {host} using IP address {addr} successful; proceeding with subsequent "
                    "IPs if any"
                ),
                host=self.args.host,
                addr=addr,
            )
        else:
            # The health check failed, log and continue on to next IP address if defined
            self.log.bind(event="error").warning(
                (
                    "Health check for {host} using IP address {addr} failed; health check will proceed with "
                    "subsequent IPs if any"
                ),
                host=self.args.host,
                addr=addr,
            )

        # Return empty result
        return None

    @abstractmethod
    def check_one(self, addr: str) -> CheckResult:
        """
        Function that is called to health check each individual IP address
        """

    def _dns_resolver(
        self, host: str, family: Literal["ipv4", "ipv6", None] = None
    ) -> list[str]:
        """Perform the requested DNS resolution. This class is called by the resolver() method and should not be used
        directly.

        Args:
            host (str): The host name or IP address.
            family (Literal["ipv4", "ipv6", None], optional): Force to the specified address family. Defaults to None.
            one (bool, optional): Only return a single IP address even if multiple IPs are available. Defaults to False.

        Raises:
            DNSResolutionError: Exception raised if the host cannot be resolved to an IP address.
            AddressFamilyError: Exception raised if the host resolves to an IP address of the wrong address family.

        Returns:
            list[str]: The list of IP address(es) of the specified address family.
        """

        # Check if there is an address family defined
        if family:
            # Address family is defined, lookup from enum
            address_family = self.AddressFamilies(family).value
        else:
            # Address family is not defined, get from init
            address_family = self.address_family

        # Set the address family type to pass to getaddrinfo
        self.log.bind(event="debug").debug(
            "Attempting to resolve host '{host}' to an IP address",
            host=host,
        )

        # Attempt the DNS resolution
        try:
            addrinfo = socket.getaddrinfo(host=host, port=0, family=address_family)
        except socket.gaierror as exc:
            # Get the type of exception based on error number
            match exc.errno:
                case "-9":
                    # No IP address of the provided address family could be found
                    message = f"Could not resolve an {family} family IP address for hostname '{host}'"
                    self.log.bind(event="error").warning(message)
                    raise AddressFamilyError(message) from exc
                case "-2":
                    message = f"Could not resolve host '{host}'; check if it exists"
                case _:
                    message = f"Could not resolve host '{host}' to an IP address: {exc}"

            # Log the failure
            self.log.bind(event="error").warning(message)

            # Raise the DNS resolution error
            raise DNSResolutionError(message=message) from exc

        # Retrieve the list of IP addresses from the addrinfo as a set to remove duplicates
        addresses = set(ip for ip in [ip[4][0] for ip in addrinfo])

        # Log the DNS resolution information
        self.log.bind(event="datadump").trace(
            "Hostname '{host}' resolved to the following IP addresses:\n{addresses}",
            host=host,
            addresses=pformat(addresses, indent=4, width=120),
        )

        # Return the resolved addresses as a list
        return list(addresses)

    def resolve(  # NOSONAR
        self, host: str, family: Literal["ipv4", "ipv6", None] = None, one: bool = False
    ) -> list[str] | str:
        """Resolve the supplied hostname into one or more IP addresses. If required, the address family may be defined
        to only return addresses of that family.

        If called with an IP address as the host argument, the IP address will be verified to be of the correct address
        family.

        Hostnames that resolve to multiple IP addresses will return a list of IP addresses unless the one argument is
        set to True, in which case only the first IP address will be returned.

        Args:
            host (str): The host name or IP address.
            family (Literal["ipv4", "ipv6", None], optional): Force to the specified address family. Defaults to None.
            one (bool, optional): Only return a single IP address even if multiple IPs are available. Defaults to False.

        Raises:
            DNSResolutionError: Exception raised if the host cannot be resolved to an IP address.
            AddressFamilyError: Exception raised if the host resolves to an IP address of the wrong address family.

        Returns:
            list[str] | str: The list of IP addresses or a single IP address.
        """
        # Perform the DNS resolution
        addresses = self._dns_resolver(host=host, family=family)

        # If a list of IP addresses is requested, return the list
        if not one:
            return addresses

        # Log that a single IP was requested
        self.log.bind(event="debug").debug(
            "Single IP address from DNS resolution requested; filtering IPs"
        )

        # If a single IP address was returned, pop and return
        if len(addresses) == 1:
            return addresses.pop()

        # Test if an address family was defined; if one was then pop the first address
        if family or hasattr(self.args, "address_family"):
            address = addresses.pop()
            self.log.bind(event="info").warning(
                "Hostname '{host}' resolved to more than one IP; only the IP '{address}' will be used",
                host=host,
                address=address,
            )
            return address

        # No address family defined; return the first IPv4 address if available
        for ipaddr in addresses:
            if isinstance(ip_address(ipaddr), IPv4Address):
                self.log.bind(event="info").warning(
                    "Hostname '{host}' resolved to more than one IP; only the first IPv4 IP '{address}' will be used",
                    host=host,
                    address=ipaddr,
                )
                return ipaddr

        # Return first IPv6 address
        address = addresses.pop()
        self.log.bind(event="info").warning(
            "Hostname '{host}' resolved to more than one IP; only the first IPv6 IP '{address}' will be used",
            host=host,
            address=address,
        )
        return address



# -*- coding: utf-8 -*-

"""
ExaCheck - ExaBGP Health Checker

Store configuration for logging to a file and/or syslog
"""

from __future__ import annotations

from os import W_OK, access
from pathlib import Path
from typing import Literal, Union

from pydantic import Field, FilePath, NewPath, field_validator, PositiveInt

from ._base import Base


class LogFile(Base):
    """
    Define and store parsed configuration for file based logging
    """

    method: Literal["file"] = Field(
        title="Log Method",
        description="Enable file based logging",
    )

    structured: bool = Field(
        title="Structured Logging",
        description="Use structured logging format",
        default=False,
    )

    destination: Union[NewPath, FilePath] = Field(
        title="Destination File",
        description="The path to the log file to write to",
        default=Path("/tmp/exacheck.log"),
    )

    size: str = Field(
        title="Log File Size",
        description="The maximum size of the log before it is rotated or overwritten",
        default="10MB",
        pattern=r"(?i)^\d+((k|m|g|t)b|b)?$",
    )

    count: PositiveInt = Field(
        title="Log Count",
        description="The number of log file rotations to keep",
        default=5,
    )

    compress: bool = Field(
        title="Compress Logs",
        description="Enable compression for the rotated logs using compression_format",
        default=True,
    )

    compression_format: Literal[
        "gz", "bz2", "xz", "lzma", "tar", "tar.gz", "tar.bz2", "tar.xz", "zip"
    ] = Field(
        title="Compression Format",
        description="If compress is enabled, the log compression format to use for rotated logs",
        default="gz",
    )

    @field_validator("destination")
    # pylint: disable=no-self-argument
    def validate_writeable(cls, destination: FilePath) -> FilePath:
        """Validate that the path to the log file/log file itself is writeable.

        Args:
            path (Path): The path to the log file

        Returns:
            Path: The original path
        """
        # Check if the log file exists already
        if destination.exists():
            # The log exists, check the log file is writable
            if not access(destination, W_OK):
                raise ValueError(
                    f"Log file '{destination}' is not writable; logs cannot be written"
                )

        # Log does not exist, check if the parent path is writable
        if not access(destination.parent, W_OK):
            raise ValueError(
                f"Log file '{destination}' parent directory '{destination.parent}' is not writable; "
                "logs cannot be written"
            )

        # Return the file path
        return destination
